import json
import logging
import os
import random
import re
import subprocess
import time

from behave import given, then, when

root_project_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..', '..'))
tests_features_path = os.path.join(root_project_path, 'tests', 'features')
test_images_path = os.path.join(root_project_path, 'tests', 'test_images')

# Import behavex-images functionality - fail if required
def get_image_attachments_module():
    """Get the image_attachments module, raising an error if not available."""
    try:
        from behavex_images import image_attachments
        return image_attachments
    except (ImportError, ModuleNotFoundError) as e:
        raise AssertionError(f"behavex-images is required for image attachment tests but not available: {e}")

# Check if behavex-images is available for optional usage
try:
    from behavex_images import image_attachments
    behavex_images_available = True
except (ImportError, ModuleNotFoundError) as e:
    behavex_images_available = False
    logging.warning(f"behavex-images not available, image attachment tests will be skipped: {e}")


@given('The progress bar is enabled')
def given_progress_bar_enabled(context):
    context.progress_bar = True


@when('I run the behavex command with a passing test')
@when('I run the behavex command with passing tests')
def when_run_passing_test(context):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex', os.path.join(tests_features_path, 'secondary_features', 'passing_tests.feature'), '-o', context.output_path]
    execute_command(context, execution_args)


@when('I run the behavex command with a file of failing tests')
def when_run_failing_tests_file(context):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex', '-rf', os.path.join(tests_features_path, 'failing_scenarios.txt'), '-o', context.output_path]
    execute_command(context, execution_args)


@when('I run the behavex command that renames scenarios and features')
def when_run_renaming_test(context):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex', os.path.join(tests_features_path, 'secondary_features', 'rename_tests.feature'), '-o', context.output_path]
    execute_command(context, execution_args)


@when('I run the behavex command with a failing test')
def when_run_failing_test(context):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex', os.path.join(tests_features_path, 'secondary_features', 'failing_tests.feature'), '-o', context.output_path]
    execute_command(context, execution_args)


@when('I run the behavex command with a crashing test')
def when_run_crashing_test(context, parallel_processes="1", parallel_scheme='scenario'):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex',
                os.path.join(tests_features_path, os.path.join(tests_features_path, 'crashing_features', 'crashing_tests.feature')),
                '-o', context.output_path]
    execute_command(context, execution_args)


@when('I run the behavex command with no tests')
def when_run_no_tests(context, parallel_processes="1", parallel_scheme='scenario'):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex',
                os.path.join(tests_features_path, os.path.join(tests_features_path, 'crashing_features', 'crashing_tests.feature')),
                '-t', 'NO_TESTS',
                '-o', context.output_path]
    execute_command(context, execution_args)


@when('I run the behavex command with a test that crashes in "{behave_hook}" hook')
@when('I run the behavex command with a test that crashes in "{behave_hook}" hook with "{parallel_processes}" parallel processes and "{parallel_scheme}" parallel scheme')
def when_run_crashing_hook_test(context, behave_hook, parallel_processes="1", parallel_scheme='scenario'):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex',
                os.path.join(tests_features_path, os.path.join(tests_features_path, 'crashing_features', 'crashing_environment_tests.feature')),
                '-o', context.output_path,
                '--parallel-processes', parallel_processes,
                '--parallel-scheme', parallel_scheme,
                '-D', 'crashing_behave_hook=' + behave_hook]
    execute_command(context, execution_args)


@when('I run the behavex command with a skipped test')
def when_run_skipped_test(context):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex', os.path.join(tests_features_path, 'secondary_features', 'skipped_tests.feature'), '-o', context.output_path]
    execute_command(context, execution_args)


@when('I run the behavex command with an untested test')
def when_run_untested_test(context):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex', os.path.join(tests_features_path, 'secondary_features', 'untested_tests.feature'), '-o', context.output_path]
    execute_command(context, execution_args)


@when('I run the behavex command with "{parallel_processes}" parallel processes and parallel scheme set as "{parallel_scheme}"')
def when_run_with_parallel_config(context, parallel_processes, parallel_scheme):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex', os.path.join(tests_features_path, 'secondary_features'), '-o', context.output_path, '--parallel-processes', parallel_processes, '--parallel-scheme', parallel_scheme]
    execute_command(context, execution_args)


@when('I setup the behavex command with "{parallel_processes}" parallel processes and parallel scheme set as "{parallel_scheme}"')
def when_setup_parallel_config(context, parallel_processes, parallel_scheme):
    context.parallel_processes = parallel_processes
    context.parallel_scheme = parallel_scheme


@when('I run the behavex command with the following scheme, processes and tags')
@when('I run the behavex command with scenario name "{scenario_name}" and the following scheme, processes and tags')
@when('I run the behavex command using "{argument_separator}" separator with the following scheme, processes and tags')
@when('I run the behavex command using "{argument_separator}" separator for "{feature_name}" feature with the following scheme, processes and tags')
@when('I run the behavex command using "{argument_separator}" separator for "{feature_name}" and "{feature_name_2}" features with the following scheme, processes and tags')
def run_command_with_scheme_processes_and_tags(context, scenario_name=None, argument_separator="equal", feature_name=None, feature_name_2=None):
    scheme = context.table[0]['parallel_scheme']
    processes = context.table[0]['parallel_processes']
    tags = context.table[0]['tags']
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    tags_to_folder_name = get_tags_string(tags)
    if not tags:
        tags_array = []
    else:
        tags_array = get_tags_arguments(tags)
    feature_path_2 = None
    if feature_name:
        if feature_name_2:
            feature_path = os.path.join(tests_features_path, 'secondary_features', feature_name)
            feature_path_2 = os.path.join(tests_features_path, 'secondary_features', feature_name_2)
        else:
            feature_path = os.path.join(tests_features_path, 'secondary_features', feature_name)
    else:
        feature_path = os.path.join(tests_features_path, 'secondary_features')
    if argument_separator == 'equal':
        execution_args = ['behavex', feature_path, '-o', context.output_path, '--parallel-processes=' + processes, '--parallel-scheme=' + scheme] + tags_array
    else:
        execution_args = ['behavex', feature_path, '-o', context.output_path, '--parallel-processes', processes, '--parallel-scheme', scheme] + tags_array
    if feature_name_2:
        # append the second feature path to the execution arguments in index 2
        execution_args.insert(2, feature_path_2)
    if scenario_name:
        execution_args.append('--name')
        execution_args.append(scenario_name.replace(' ', '\\ '))
    execute_command(context, execution_args)


@when('I run the behavex command with the following tags')
def when_run_with_tags(context):
    tags = context.table[0]['tags']
    tags_to_folder_name = get_tags_string(tags)
    tags_array = get_tags_arguments(tags)
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex', os.path.join(tests_features_path, 'secondary_features'), '-o', context.output_path] + tags_array
    execute_command(context, execution_args)


@when('I run the behavex command by performing a dry run')
def when_run_dry_run(context):
    # generate a random number between 1 and 1000000 completing with zeroes to 6 digits
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex', os.path.join(tests_features_path, 'secondary_features'), '-o', context.output_path, '--dry-run']
    execute_command(context, execution_args)


@then('I should see the following behavex console outputs')
@then('I should see the following behavex console outputs and exit code "{expected_exit_code}"')
def then_see_console_outputs(context, expected_exit_code=None):
    if expected_exit_code is not None:
        assert int(context.result.returncode) == int(expected_exit_code), "Behavex exit code is not expected"
    for row in context.table:
        assert row['output_line'] in context.result.stdout, f"Unexpected output when checking console outputs: {context.result.stdout}\n\nOutput line not found: {row['output_line']}\n"


@then('I should not see error messages in the output')
def then_no_error_messages(context):
    error_messages = ["error", "exception", "traceback"]
    # Filter out expected behavex-images errors during dry run
    output_lines = context.result.stdout.split('\n')
    filtered_output = []

    for line in output_lines:
        # Skip expected behavex-images errors during dry run
        if '[behavex-images]' in line and ('expected str, bytes or os.PathLike object, not NoneType' in line or
                                           'It was not possible to add the image to the report' in line):
            continue
        # Skip expected error messages from test scenarios during dry run
        if ('ERROR - Executing permanently broken action' in line or
            'ERROR - This step is designed to fail' in line or
            'ERROR - Permanently broken action failed' in line or
            'ERROR - Flaky action failed (expected' in line or
            'ERROR - Action failed without retry (expected' in line or
            'ERROR - Unexpected output when checking error messages in the console output:' in line or
            'BehaveX AUTO-RETRY: Scenario' in line or
            'image_attachments_undefined_step.feature:' in line or
            'Errored scenarios:' in line or
            'Failing scenarios:' in line or
            ('features passed' in line and 'failed' in line and 'error' in line) or  # Summary lines like "X features passed, Y failed, Z error"
            ('scenarios passed' in line and 'failed' in line) or  # Summary lines like "X scenarios passed, Y failed"
            ('steps passed' in line and 'failed' in line)):  # Summary lines like "X steps passed, Y failed"
            continue
        filtered_output.append(line)

    filtered_output_text = '\n'.join(filtered_output)

    for message in error_messages:
        assert message not in filtered_output_text.lower(), f"Unexpected output when checking error messages in the console output: {context.result.stdout}\n"


@then('I should not see exception messages in the output')
def then_no_exception_messages(context):
    exception_messages = ["exception", "traceback"]
    for message in exception_messages:
        assert message not in context.result.stdout.lower(), f"Unexpected output when checking exception messages in the console output: {context.result.stdout}\n"





@then('I should see the same number of scenarios in the reports and the console output')
def then_same_scenarios_count_all(context):
    total_scenarios_in_html_report = get_total_scenarios_in_html_report(context)
    logging.info(f"Total scenarios in the HTML report: {total_scenarios_in_html_report}")
    total_scenarios_in_junit_reports = get_total_scenarios_in_junit_reports(context)
    logging.info(f"Total scenarios in the JUnit reports: {total_scenarios_in_junit_reports}")
    total_scenarios_in_console_output = get_total_scenarios_in_console_output(context)
    logging.info(f"Total scenarios in the console output: {total_scenarios_in_console_output}")
    assert total_scenarios_in_html_report == total_scenarios_in_junit_reports == total_scenarios_in_console_output, f"Expected total scenarios to match, but found {total_scenarios_in_html_report} in the HTML report, {total_scenarios_in_junit_reports} in the JUnit reports, and {total_scenarios_in_console_output} in the console output"


@then('I should see the same number of scenarios in the reports')
def verify_total_scenarios_in_reports(context, consider_skipped_scenarios=True):
    total_scenarios_in_html_report = get_total_scenarios_in_html_report(context)
    logging.info(f"Total scenarios in the HTML report: {total_scenarios_in_html_report}")
    total_scenarios_in_junit_reports = get_total_scenarios_in_junit_reports(context, consider_skipped_scenarios)
    logging.info(f"Total scenarios in the JUnit reports: {total_scenarios_in_junit_reports}")
    assert total_scenarios_in_html_report == total_scenarios_in_junit_reports, f"Expected total scenarios to match, but found {total_scenarios_in_html_report} in the HTML report, {total_scenarios_in_junit_reports} in the JUnit reports"


@then('I should see the same number of scenarios in the reports not considering the skipped scenarios')
def then_same_scenarios_count_excluding_skipped(context):
    verify_total_scenarios_in_reports(context, consider_skipped_scenarios=False)


@then('I should see the HTML report was generated')
def verify_html_report_exists(context):
    """Verify that the HTML report file exists and is a valid file."""
    report_path = os.path.abspath(os.path.join(context.output_path, 'report.html'))
    assert os.path.exists(report_path), f"Expected HTML report to exist at {report_path}"
    assert os.path.getsize(report_path) > 0, f"Expected HTML report to be non-empty at {report_path}"


@then('I should see the HTML report was generated and contains scenarios')
@then('I should see the HTML report was generated and contains "{total_scenarios}" scenarios')
def verify_total_scenarios_in_html_report(context, total_scenarios=None, consider_skipped_scenarios=True):
    total_scenarios_in_html_report = get_total_scenarios_in_html_report(context)
    logging.info(f"Total scenarios in the HTML report: {total_scenarios_in_html_report}")
    if total_scenarios is not None:
        assert total_scenarios_in_html_report == int(total_scenarios), f"Expected the HTML report to contain {total_scenarios} scenarios, but found {total_scenarios_in_html_report}"
    else:
        assert total_scenarios_in_html_report > 0, "Expected the HTML report to be generated and contain scenarios"


@then('I should see the generated HTML report contains the "{string_to_search}" string')
def verify_string_in_html_report(context, string_to_search, string_should_be_present=True):
    total_string_instances_in_html_report = get_string_instances_from_html_report(context, string_to_search)
    logging.info(f"Total instances of '{string_to_search}' in the HTML report: {total_string_instances_in_html_report}")
    if string_should_be_present:
        assert total_string_instances_in_html_report > 0, f"Expected the HTML report to contain the string '{string_to_search}'"
    else:
        assert total_string_instances_in_html_report == 0, f"Expected the HTML report to not contain the string '{string_to_search}'"


@then('I should see the generated HTML report does not contain internal BehaveX variables and tags')
def verify_string_not_in_html_report(context):
    internal_behavex_variables_and_tags = ["BHX_", "BHX_TAG_"]
    for variable_or_tag in internal_behavex_variables_and_tags:
        total_string_instances_in_html_report = get_string_instances_from_html_report(context, variable_or_tag)
        logging.info(f"Total instances of '{variable_or_tag}' in the HTML report: {total_string_instances_in_html_report}")
        assert total_string_instances_in_html_report == 0, f"Expected the HTML report to not contain the string '{variable_or_tag}'"



def get_tags_arguments(tags):
    tags_array = []
    for tag in tags.split(' '):
        tags_array += tag.split('=')
    return tags_array


def get_tags_string(tags):
    return tags.replace('-t=','_AND_').replace('~','NOT_').replace(',','_OR_').replace(' ','').replace('@','')


def get_random_number(total_digits):
    return str(random.randint(1, 1000000)).zfill(total_digits)


def get_total_scenarios_in_console_output(context):
    #Verifying the scenarios in the console output
    console_output = context.result.stdout
    # Extract the number of scenarios by analyzing the following pattern: X scenarios passed, Y failed, Z skipped
    scenario_pattern = re.compile(r'(\d+) scenario.? passed, (\d+) failed, (\d+) skipped')
    match = scenario_pattern.search(console_output)
    if match:
        scenarios_passed = int(match.group(1))
        scenarios_failed = int(match.group(2))
        scenarios_skipped = int(match.group(3))
    else:
        raise ValueError("No scenarios found in the console output")
    return scenarios_passed + scenarios_failed + scenarios_skipped


def get_total_scenarios_in_html_report(context):
    report_path = os.path.abspath(os.path.join(context.output_path, 'report.html'))
    with open(report_path, 'r') as file:
        html_content = file.read()
    return html_content.count('data-scenario-tags=')


def get_string_instances_from_html_report(context, string_to_search):
    report_path = os.path.abspath(os.path.join(context.output_path, 'report.html'))
    with open(report_path, 'r') as file:
        html_content = file.read()
    return html_content.lower().count(string_to_search.lower())


def get_total_scenarios_in_junit_reports(context, consider_skipped_scenarios=True):
    junit_folder = os.path.abspath(os.path.join(context.output_path, 'behave'))
    total_scenarios_in_junit_reports = 0
    for file in os.listdir(junit_folder):
        if file.endswith('.xml'):
            with open(os.path.join(junit_folder, file), 'r') as file:
                xml_content = file.read()
                total_scenarios_in_junit_reports += xml_content.count('status="passed"')
                total_scenarios_in_junit_reports += xml_content.count('status="failed"')
                # Keep status="error" counting for robustness, but shouldn't be needed now
                total_scenarios_in_junit_reports += xml_content.count('status="error"')
                if consider_skipped_scenarios:
                    total_scenarios_in_junit_reports += xml_content.count('status="skipped"')
    return total_scenarios_in_junit_reports


@given('I take a screenshot using test image {image_number}')
def step_take_screenshot_using_test_image(context, image_number):
    """Attach a test image file using behavex-images"""
    image_path = os.path.join(test_images_path, f'test_image_{image_number}.png')
    if not os.path.exists(image_path):
        raise FileNotFoundError(f"Test image not found: {image_path}")

    # Check if image attachments are required for this test
    if hasattr(context, 'require_image_attachments') and context.require_image_attachments:
        image_attachments = get_image_attachments_module()
        image_attachments.attach_image_file(context, image_path)
        logging.info(f"Attached test image: {image_path}")
    else:
        # Optional usage - skip if not available
        if not behavex_images_available:
            logging.warning("behavex-images not available, skipping image attachment")
            return
        # Import here to avoid issues if behavex-images is not available
        from behavex_images import image_attachments
        image_attachments.attach_image_file(context, image_path)
        logging.info(f"Attached test image: {image_path}")


@given('image attachments are configured for ONLY_ON_FAILURE condition')
def step_configure_only_on_failure_attachments(context):
    """Configure image attachments to use ONLY_ON_FAILURE condition"""
    context.attachment_condition = 'ONLY_ON_FAILURE'

@when('I run the behavex command with image attachments')
def step_run_behavex_with_image_attachments(context):
    """Run behavex with HTML formatter and image attachments enabled"""
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = [
        'behavex',
        os.path.join(tests_features_path, 'secondary_features', 'simple_image_tests.feature'),
        '-o', context.output_path,
        '-t', '@IMAGE_ATTACHMENT',
        '-t', '@HTML_REPORT'
    ]
    execute_command(context, execution_args)


@when('I run the behavex command with failed scenario and image attachments')
def step_run_behavex_with_failed_scenario_and_image_attachments(context):
    """Run behavex with failed scenario and image attachments enabled"""
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = [
        'behavex',
        os.path.join(tests_features_path, 'secondary_features', 'image_attachments_failing_step.feature'),
        '-o', context.output_path,
        '-t', '@IMAGE_ATTACHMENT',
        '-t', '@FAILED_SCENARIO'
    ]
    # Expect non-zero exit code due to failed scenario
    context.expected_exit_code = 1
    execute_command(context, execution_args)

@when('I run the behavex command with error scenario and image attachments')
def step_run_behavex_with_error_scenario_and_image_attachments(context):
    """Run behavex with error scenario and image attachments enabled"""
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = [
        'behavex',
        os.path.join(tests_features_path, 'secondary_features', 'image_attachments_undefined_step.feature'),
        '-o', context.output_path,
        '-t', '@IMAGE_ATTACHMENT',
        '-t', '@ERROR_SCENARIO'
    ]
    # Expect non-zero exit code due to error scenario
    context.expected_exit_code = 1
    execute_command(context, execution_args)


@then('I should see that images are attached for error scenarios in the HTML report')
def step_verify_images_attached_for_error_scenarios(context):
    """Verify that images are attached in HTML report for scenarios with error status"""
    html_report_path = os.path.join(context.output_path, 'report.html')
    assert os.path.exists(html_report_path), f"HTML report not found at {html_report_path}"

    with open(html_report_path, 'r', encoding='utf-8') as f:
        html_content = f.read()

    # Verify that the HTML contains image attachments for error scenarios
    assert "glyphicon-picture" in html_content, "Image attachment icon not found in HTML report"

    # Look for specific patterns that indicate images are attached to error scenarios
    # The HTML should contain scenario rows with error status and image attachments
    assert "error" in html_content.lower(), "Error scenario not found in HTML report"

    logging.info("Successfully verified that images are attached for error scenarios in HTML report")


@then('I should see that images are attached for failed scenarios in the HTML report')
def step_verify_images_attached_for_failed_scenarios(context):
    """Verify that images are attached in HTML report for scenarios with failed status"""
    html_report_path = os.path.join(context.output_path, 'report.html')
    assert os.path.exists(html_report_path), f"HTML report not found at {html_report_path}"

    with open(html_report_path, 'r', encoding='utf-8') as f:
        html_content = f.read()

    # Verify that the HTML contains image attachments for failed scenarios
    assert "glyphicon-picture" in html_content, "Image attachment icon not found in HTML report"

    # Look for specific patterns that indicate images are attached to failed scenarios
    # The HTML should contain scenario rows with failed status and image attachments
    assert "failed" in html_content.lower(), "Failed scenario not found in HTML report"

    logging.info("Successfully verified that images are attached for failed scenarios in HTML report")


@then('I should verify current behavior for error scenarios with ONLY_ON_FAILURE condition')
def step_verify_current_behavior_error_scenarios(context):
    """Verify current behavior for error scenarios with ONLY_ON_FAILURE condition"""
    html_report_path = os.path.join(context.output_path, 'report.html')
    assert os.path.exists(html_report_path), f"HTML report not found at {html_report_path}"

    with open(html_report_path, 'r', encoding='utf-8') as f:
        html_content = f.read()

    # Verify that the error scenario exists in the report
    assert "error" in html_content.lower(), "Error scenario not found in HTML report"

    # Check current behavior - as of now, ONLY_ON_FAILURE might not attach images for error status
    # This documents the current behavior rather than asserting what should happen
    if "glyphicon-picture" in html_content:
        logging.info("Current behavior: Images ARE attached for error scenarios with ONLY_ON_FAILURE condition")
    else:
        logging.info("Current behavior: Images are NOT attached for error scenarios with ONLY_ON_FAILURE condition")
        logging.info("This may be a feature gap - error status should be treated similar to failed status")

    # The test passes regardless to document current behavior
    logging.info("Successfully documented current behavior for error scenarios with ONLY_ON_FAILURE condition")


@when('I run the behavex command with allure formatter and image attachments')
def step_run_behavex_with_allure_and_image_attachments(context):
    """Run behavex with Allure formatter and image attachments enabled"""
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = [
        'behavex',
        os.path.join(tests_features_path, 'secondary_features', 'simple_image_tests.feature'),
        '-o', context.output_path,
        '--formatter=behavex.outputs.formatters.allure_behavex_formatter:AllureBehaveXFormatter',
        '-t', '@IMAGE_ATTACHMENT',
        '-t', '@ALLURE_FORMATTER'
    ]
    execute_command(context, execution_args)


@then('I should see the HTML report contains screenshots')
def step_verify_html_report_contains_screenshots(context):
    """Verify that the HTML report contains attached screenshots"""
    html_report_path = os.path.join(context.output_path, 'report.html')
    if os.path.exists(html_report_path):
        with open(html_report_path, 'r', encoding='utf-8') as f:
            html_content = f.read()
            # Look for image references in the HTML
            assert 'test_image_' in html_content, "Screenshots not found in HTML report"
            logging.info("Screenshots found in HTML report")
    else:
        raise FileNotFoundError(f"HTML report not found: {html_report_path}")


@then('I should see the HTML report contains the image gallery icon')
def step_verify_html_report_contains_gallery_icon(context):
    """Verify that the HTML report contains the image gallery icon"""
    html_report_path = os.path.join(context.output_path, 'report.html')
    if os.path.exists(html_report_path):
        with open(html_report_path, 'r', encoding='utf-8') as f:
            html_content = f.read()
            # Look for the specific gallery icon HTML element
            gallery_icon = '<span class="glyphicon glyphicon-picture">'
            assert gallery_icon in html_content, "Image gallery icon not found in HTML report"
            logging.info("Image gallery icon found in HTML report")
    else:
        raise FileNotFoundError(f"HTML report not found: {html_report_path}")


@then('I should see that images are stored in the allure formatter output directory')
def step_verify_images_in_allure_output(context):
    """Verify that images are stored in the Allure formatter output directory"""
    allure_results_path = os.path.join(context.output_path, 'allure-results')
    if os.path.exists(allure_results_path):
        # Look for image attachments in allure results
        allure_files = os.listdir(allure_results_path)
        image_files = [f for f in allure_files if f.endswith('.png') or f.endswith('.jpg') or f.endswith('.jpeg')]
        assert len(image_files) > 0, "No image files found in Allure output directory"
        logging.info(f"Found {len(image_files)} image files in Allure output directory")

        # Also check for attachment references in JSON files
        json_files = [f for f in allure_files if f.endswith('.json')]
        found_attachment_reference = False
        for json_file in json_files:
            json_path = os.path.join(allure_results_path, json_file)
            with open(json_path, 'r', encoding='utf-8') as f:
                json_content = f.read()
                if 'attachments' in json_content and 'image' in json_content:
                    found_attachment_reference = True
                    break

        assert found_attachment_reference, "No image attachment references found in Allure JSON files"
        logging.info("Image attachment references found in Allure JSON files")
    else:
        raise FileNotFoundError(f"Allure results directory not found: {allure_results_path}")


def execute_command(context, execution_args, print_output=True):
    if "progress_bar" in context and context.progress_bar:
        execution_args.insert(2, '--show-progress-bar')
    if hasattr(context, 'parallel_processes'):
        execution_args += ['--parallel-processes', context.parallel_processes]
    if hasattr(context, 'parallel_scheme'):
        execution_args += ['--parallel-scheme', context.parallel_scheme]
    logging.info("Executing command: {}".format(" ".join(execution_args)))
    context.result = subprocess.run(execution_args, capture_output=True, text=True)
    if print_output:
        logging.info(context.result.stdout)


# ---------- Execution Ordering Test Steps ----------

@when('I run the behavex command with execution ordering enabled using "{parallel_processes}" parallel processes and parallel scheme set as "{parallel_scheme}"')
def when_run_with_execution_ordering(context, parallel_processes, parallel_scheme):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex',
                      os.path.join(tests_features_path, 'secondary_features', 'ordered_tests.feature'),
                      '-t', '@ORDERED_TEST',
                      '-o', context.output_path,
                      '--parallel-processes', parallel_processes,
                      '--parallel-scheme', parallel_scheme,
                      '--order-tests']
    execute_command(context, execution_args)


@when('I run the behavex command with execution ordering enabled using custom prefix "{prefix}" with "{parallel_processes}" parallel processes and parallel scheme set as "{parallel_scheme}"')
def when_run_with_custom_prefix_ordering(context, prefix, parallel_processes, parallel_scheme):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex',
                      os.path.join(tests_features_path, 'secondary_features', 'priority_ordered_tests.feature'),
                      '-t', '@PRIORITY_TEST',
                      '-o', context.output_path,
                      '--parallel-processes', parallel_processes,
                      '--parallel-scheme', parallel_scheme,
                      '--order-tests',
                      '--order-tag-prefix', prefix]
    execute_command(context, execution_args)


@when('I run the behavex command with execution ordering enabled for mixed scenarios using "{parallel_processes}" parallel processes and parallel scheme set as "{parallel_scheme}"')
def when_run_mixed_ordering(context, parallel_processes, parallel_scheme):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex',
                      os.path.join(tests_features_path, 'secondary_features', 'mixed_ordered_tests.feature'),
                      '-t', '@MIXED_TEST',
                      '-o', context.output_path,
                      '--parallel-processes', parallel_processes,
                      '--parallel-scheme', parallel_scheme,
                      '--order-tests']
    execute_command(context, execution_args)


@when('I run the behavex command with strict execution ordering enabled using "{parallel_processes}" parallel processes and parallel scheme set as "{parallel_scheme}"')
def when_run_strict_ordering(context, parallel_processes, parallel_scheme):
    context.output_path = os.path.join('output', 'output_{}'.format(get_random_number(6)))
    execution_args = ['behavex',
                      os.path.join(tests_features_path, 'secondary_features', 'ordered_tests.feature'),
                      '-t', '@ORDERED_TEST',
                      '-o', context.output_path,
                      '--parallel-processes', parallel_processes,
                      '--parallel-scheme', parallel_scheme,
                      '--order-tests-strict']
    execute_command(context, execution_args)


@then('I should see the scenarios executed in the correct order for "{parallel_scheme}" scheme')
def then_see_correct_execution_order(context, parallel_scheme):
    """Verify that scenarios with ORDER tags executed successfully (--order-tests controls submission order, not start time)"""
    report_json_path = os.path.join(context.output_path, 'report.json')
    assert os.path.exists(report_json_path), f"Report JSON file not found at {report_json_path}"

    with open(report_json_path, 'r') as json_file:
        report_data = json.load(json_file)

    # Extract scenario execution data with order values and start times
    scenario_order_mapping = {
        "First ordered test scenario": 1,
        "Second ordered test scenario": 2,
        "Third ordered test scenario": 3,
        "Fourth ordered test scenario": 10
    }

    executed_scenarios = []
    for feature in report_data.get('features', []):
        for scenario in feature.get('scenarios', []):
            if scenario['name'] in scenario_order_mapping:
                # Ensure scenario has start time
                assert 'start' in scenario, f"Scenario '{scenario['name']}' missing start time in report"

                executed_scenarios.append({
                    'name': scenario['name'],
                    'order': scenario_order_mapping[scenario['name']],
                    'start_time': scenario['start']
                })

    # Verify that all expected scenarios were executed
    expected_count = 4
    assert len(executed_scenarios) == expected_count, \
        f"Expected {expected_count} scenarios to be executed, but got {len(executed_scenarios)}"

    # Sort by actual start time for logging purposes
    executed_scenarios.sort(key=lambda x: x['start_time'])

    # Log the actual execution order with start times for informational purposes
    logging.info("📊 Scenario execution order (--order-tests controls submission order, not start time):")
    for i, scenario in enumerate(executed_scenarios, 1):
        logging.info(f"  {i}. ORDER_{scenario['order']:03d}: {scenario['name']} started at {scenario['start_time']}")

    # For --order-tests: We only verify that scenarios were submitted in order (not start time)
    # Since submission order cannot be verified from JSON report, we accept any start time order
    # and only validate that all scenarios executed successfully

    # Sort by order value to show intended vs actual execution order
    expected_order = sorted(executed_scenarios, key=lambda x: x['order'])
    logging.info("📋 Expected submission order vs actual start time order:")

    expected_order_str = " → ".join([f"ORDER_{s['order']:03d}" for s in expected_order])
    actual_order_str = " → ".join([f"ORDER_{s['order']:03d}" for s in executed_scenarios])

    logging.info(f"  Expected: {expected_order_str}")
    logging.info(f"  Actual:   {actual_order_str}")

    if [s['order'] for s in expected_order] != [s['order'] for s in executed_scenarios]:
        logging.info("ℹ️  Start time order differs from submission order - this is normal for --order-tests")
        logging.info("   --order-tests controls submission order, allowing parallel execution with timing variations")
    else:
        logging.info("✅ Start time order matches submission order (optimal parallel scheduling)")

    logging.info(f"✅ Order tests validation passed: All {expected_count} scenarios executed successfully")


@then('I should see the scenarios executed in the correct order for "{parallel_scheme}" scheme with custom prefix')
def then_see_correct_order_custom_prefix(context, parallel_scheme):
    """Verify that scenarios with PRIORITY tags executed successfully (--order-tests controls submission order, not start time)"""
    report_json_path = os.path.join(context.output_path, 'report.json')
    assert os.path.exists(report_json_path), f"Report JSON file not found at {report_json_path}"

    with open(report_json_path, 'r') as json_file:
        report_data = json.load(json_file)

    # Extract scenario execution data with priority values and start times
    scenario_priority_mapping = {
        "First priority test scenario": 1,
        "Second priority test scenario": 2,
        "Third priority test scenario": 3
    }

    executed_scenarios = []
    for feature in report_data.get('features', []):
        for scenario in feature.get('scenarios', []):
            if scenario['name'] in scenario_priority_mapping:
                # Ensure scenario has start time
                assert 'start' in scenario, f"Scenario '{scenario['name']}' missing start time in report"

                executed_scenarios.append({
                    'name': scenario['name'],
                    'priority': scenario_priority_mapping[scenario['name']],
                    'start_time': scenario['start']
                })

    # Verify that all expected scenarios were executed
    expected_count = 3
    assert len(executed_scenarios) == expected_count, \
        f"Expected {expected_count} scenarios to be executed, but got {len(executed_scenarios)}"

    # Sort by actual start time for logging purposes
    executed_scenarios.sort(key=lambda x: x['start_time'])

    # Log the actual execution order with start times for informational purposes
    logging.info("📊 Custom prefix scenario execution order (--order-tests controls submission order, not start time):")
    for i, scenario in enumerate(executed_scenarios, 1):
        logging.info(f"  {i}. PRIORITY_{scenario['priority']:03d}: {scenario['name']} started at {scenario['start_time']}")

    # Sort by priority value to show intended vs actual execution order
    expected_order = sorted(executed_scenarios, key=lambda x: x['priority'])
    logging.info("📋 Expected submission order vs actual start time order:")

    expected_order_str = " → ".join([f"PRIORITY_{s['priority']:03d}" for s in expected_order])
    actual_order_str = " → ".join([f"PRIORITY_{s['priority']:03d}" for s in executed_scenarios])

    logging.info(f"  Expected: {expected_order_str}")
    logging.info(f"  Actual:   {actual_order_str}")

    if [s['priority'] for s in expected_order] != [s['priority'] for s in executed_scenarios]:
        logging.info("ℹ️  Start time order differs from submission order - this is normal for --order-tests")
        logging.info("   --order-tests controls submission order, allowing parallel execution with timing variations")
    else:
        logging.info("✅ Start time order matches submission order (optimal parallel scheduling)")

    logging.info(f"✅ Custom prefix order tests validation passed: All {expected_count} scenarios executed successfully")


@then('I should see that ordered scenarios execute before unordered scenarios for "{parallel_scheme}" scheme')
def then_see_ordered_before_unordered(context, parallel_scheme):
    """Verify that ordered and unordered scenarios executed successfully (--order-tests controls submission order, not start time)"""
    report_json_path = os.path.join(context.output_path, 'report.json')
    assert os.path.exists(report_json_path), f"Report JSON file not found at {report_json_path}"

    with open(report_json_path, 'r') as json_file:
        report_data = json.load(json_file)

    # Define expected scenarios and their order values
    scenario_order_mapping = {
        "First ordered test scenario": 1,
        "Second ordered test scenario": 2,
        "Third ordered test scenario": 3,
        "Unordered test scenario that should run last": 9999,
        "Another unordered test scenario that should run last": 9999
    }

    executed_scenarios = []
    for feature in report_data.get('features', []):
        for scenario in feature.get('scenarios', []):
            if scenario['name'] in scenario_order_mapping:
                # Ensure scenario has start time
                assert 'start' in scenario, f"Scenario '{scenario['name']}' missing start time in report"

                executed_scenarios.append({
                    'name': scenario['name'],
                    'order': scenario_order_mapping[scenario['name']],
                    'start_time': scenario['start'],
                    'is_ordered': scenario_order_mapping[scenario['name']] < 9999
                })

    # Verify that all expected scenarios were executed
    expected_count = 5
    assert len(executed_scenarios) == expected_count, \
        f"Expected {expected_count} scenarios to be executed, but got {len(executed_scenarios)}"

    # Separate ordered and unordered scenarios
    ordered_scenarios = [s for s in executed_scenarios if s['is_ordered']]
    unordered_scenarios = [s for s in executed_scenarios if not s['is_ordered']]

    # Verify counts
    assert len(ordered_scenarios) == 3, f"Expected 3 ordered scenarios, got {len(ordered_scenarios)}"
    assert len(unordered_scenarios) == 2, f"Expected 2 unordered scenarios, got {len(unordered_scenarios)}"

    # Sort by start time for logging
    executed_scenarios.sort(key=lambda x: x['start_time'])

    # Log the execution order with start times for informational purposes
    logging.info("📊 Mixed scenario execution order (--order-tests controls submission order, not start time):")
    for i, scenario in enumerate(executed_scenarios, 1):
        order_type = "ORDERED" if scenario['is_ordered'] else "UNORDERED"
        order_value = f"ORDER_{scenario['order']:03d}" if scenario['is_ordered'] else "NO_ORDER"
        logging.info(f"  {i}. {order_value} ({order_type}): {scenario['name']} started at {scenario['start_time']}")

    # For --order-tests: Ordered scenarios are submitted before unordered scenarios
    # But actual start times may vary due to parallel scheduling
    logging.info("📋 Submission order expectation: All ordered scenarios submitted before unordered scenarios")
    logging.info("   Actual start times may vary due to parallel worker scheduling")

    # Check if ordered scenarios actually started before unordered ones (informational)
    latest_ordered_start = max(s['start_time'] for s in ordered_scenarios)
    earliest_unordered_start = min(s['start_time'] for s in unordered_scenarios)

    if latest_ordered_start <= earliest_unordered_start:
        logging.info("✅ Start time order matches submission order: All ordered scenarios started before unordered scenarios")
    else:
        logging.info("ℹ️  Start time order differs from submission order - this is normal for --order-tests")
        logging.info("   Some unordered scenarios started before ordered scenarios due to parallel scheduling")

    logging.info(f"✅ Mixed scenario validation passed: All ordered and unordered scenarios executed successfully")


@then('the execution order should not be enforced with single process execution')
def then_no_order_enforcement_single_process(context):
    """Verify that ordering is not enforced when running with single process"""
    # For single process execution, we just verify that the test completed successfully
    # and scenarios were executed (the actual order doesn't matter since ordering only works with parallel execution)


@then('I should see the scenarios executed in strict order for "{parallel_scheme}" scheme')
def then_see_strict_execution_order(context, parallel_scheme):
    """Verify that scenarios with ORDER tags executed in strict sequential order when using --order-tests-strict"""
    report_json_path = os.path.join(context.output_path, 'report.json')
    assert os.path.exists(report_json_path), f"Report JSON file not found at {report_json_path}"

    with open(report_json_path, 'r') as json_file:
        report_data = json.load(json_file)

    # Extract scenario execution data with order values, start times, and stop times
    scenario_order_mapping = {
        "First ordered test scenario": 1,
        "Second ordered test scenario": 2,
        "Third ordered test scenario": 3,
        "Fourth ordered test scenario": 10
    }

    executed_scenarios = []
    for feature in report_data.get('features', []):
        for scenario in feature.get('scenarios', []):
            if scenario['name'] in scenario_order_mapping:
                # Ensure scenario has both start and stop times
                assert 'start' in scenario, f"Scenario '{scenario['name']}' missing start time in report"
                assert 'stop' in scenario, f"Scenario '{scenario['name']}' missing stop time in report"

                executed_scenarios.append({
                    'name': scenario['name'],
                    'order': scenario_order_mapping[scenario['name']],
                    'start_time': scenario['start'],
                    'stop_time': scenario['stop']
                })

    # Verify that all expected scenarios were executed
    expected_count = 4
    assert len(executed_scenarios) == expected_count, \
        f"Expected {expected_count} scenarios to be executed, but got {len(executed_scenarios)}"

    # For --order-tests-strict: Verify strict sequential execution
    # Each scenario must complete before the next scenario starts
    executed_scenarios.sort(key=lambda x: x['start_time'])

    if parallel_scheme == 'scenario':
        # For scenario-level execution: scenarios should execute in ORDER tag order
        logging.info("📊 Scenario-level strict execution order verification:")
        expected_order = sorted(executed_scenarios, key=lambda x: x['order'])
        for i, scenario in enumerate(expected_order):
            actual_scenario = executed_scenarios[i]
            assert scenario['order'] == actual_scenario['order'], \
                f"Scenario execution order mismatch: expected ORDER_{scenario['order']:03d} ({scenario['name']}) " \
                f"but got ORDER_{actual_scenario['order']:03d} ({actual_scenario['name']}) at position {i+1}"
    else:
        # For feature-level execution: scenarios execute in file order (behave's natural behavior)
        # but still in strict sequential order (each must complete before next starts)
        logging.info("📊 Feature-level strict execution order verification:")
        logging.info("    Note: With feature-level execution, scenarios execute in file order, not ORDER tag order")

    # Log the execution order for debugging
    for i, scenario in enumerate(executed_scenarios, 1):
        duration = scenario['stop_time'] - scenario['start_time']
        logging.info(f"  {i}. ORDER_{scenario['order']:03d}: {scenario['name']}")
        logging.info(f"    Start: {scenario['start_time']}, Stop: {scenario['stop_time']}, Duration: {duration:.3f}s")

    # Verify strict sequential execution for both schemes
    for i in range(len(executed_scenarios) - 1):
        current_scenario = executed_scenarios[i]
        next_scenario = executed_scenarios[i + 1]

        # The critical requirement for strict ordering: previous scenario must complete before next starts
        assert current_scenario['stop_time'] <= next_scenario['start_time'], \
            f"Strict sequential execution violated: {current_scenario['name']} (ORDER_{current_scenario['order']:03d}) " \
            f"stopped at {current_scenario['stop_time']} but {next_scenario['name']} (ORDER_{next_scenario['order']:03d}) " \
            f"started at {next_scenario['start_time']}. With --order-tests-strict, scenarios must complete before the next starts!"

    logging.info(f"✅ Strict sequential execution verified for {parallel_scheme} scheme with --order-tests-strict")



# Configuration file for the Sphinx documentation builder.

import os
import sys

sys.path.insert(0, os.path.abspath('..'))

# -- Project information -----------------------------------------------------
project = 'BehaveX'
copyright = '2023, hrcorval'
author = 'hrcorval'

# -- General configuration ---------------------------------------------------
extensions = [
    'myst_parser',
    'autoapi.extension',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
]

# The master toctree document.
master_doc = 'index'

# The suffix(es) of source filenames.
source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
}

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Options for HTML output -------------------------------------------------
html_theme = 'sphinx_rtd_theme'
templates_path = ['_templates']

# -- MyST Parser configuration -----------------------------------------------
myst_enable_extensions = [
    "colon_fence",
    "deflist",
    "tasklist",
    "substitution",
    "linkify",
]

# -- AutoAPI configuration ---------------------------------------------------
autoapi_dirs = ['../behavex']
autoapi_type = 'python'
autoapi_template_dir = '_autoapi_templates'
autoapi_generate_api_docs = True
autoapi_add_toctree_entry = False  # We'll add it manually to control placement

# -- HTML output options ----------------------------------------------------
html_static_path = []
html_show_sourcelink = True
html_show_sphinx = True
html_show_copyright = True



#!/usr/bin/env python3
"""
BehaveX HTML Report Generator

Utility script to generate HTML report from existing report.json file.
This is useful when you want to regenerate the HTML report without re-running tests.

Usage: python generate_html_from_json.py [path_to_report.json] [output_directory]

Examples:
    # Generate HTML in the same directory as the JSON file
    python generate_html_from_json.py output/report.json

    # Generate HTML in a specific directory
    python generate_html_from_json.py output/report.json my_reports/

    # Works with any BehaveX JSON report
    python generate_html_from_json.py /path/to/old_execution/report.json
"""
import json
import os
import sys
import time
from pathlib import Path

# Add current directory to path so we can import behavex modules
sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))

from behavex.conf_mgr import set_env
from behavex.global_vars import global_vars
from behavex.outputs import report_html


def generate_html_from_json(json_file_path, output_dir=None):
    """Generate HTML report from existing JSON report file."""

    # Set default output directory if not provided
    if output_dir is None:
        output_dir = os.path.dirname(json_file_path) or 'output'

    # Ensure output directory exists
    os.makedirs(output_dir, exist_ok=True)

    # Set environment variables that BehaveX expects (keys are stored in lowercase)
    set_env('output', output_dir)
    set_env('temp', os.path.join(output_dir, 'temp'))
    set_env('logs', os.path.join(output_dir, 'outputs', 'logs'))

    # Create necessary directories
    os.makedirs(os.path.join(output_dir, 'outputs', 'bootstrap', 'css'), exist_ok=True)
    os.makedirs(os.path.join(output_dir, 'outputs', 'bootstrap', 'manifest'), exist_ok=True)
    os.makedirs(os.path.join(output_dir, 'outputs', 'logs'), exist_ok=True)
    os.makedirs(os.path.join(output_dir, 'temp'), exist_ok=True)

    # Load the JSON report first to get real execution times
    with open(json_file_path, 'r', encoding='utf-8') as f:
        json_data = json.load(f)

    # Calculate real execution start and end times from scenario data
    scenario_start_times = []
    scenario_end_times = []

    for feature in json_data.get('features', []):
        for scenario in feature.get('scenarios', []):
            if 'start' in scenario:
                start_time = scenario['start']
                # Convert milliseconds to seconds if needed
                if start_time > 1000000000000:  # milliseconds
                    start_time = start_time / 1000
                scenario_start_times.append(start_time)

            if 'stop' in scenario:
                stop_time = scenario['stop']
                # Convert milliseconds to seconds if needed
                if stop_time > 1000000000000:  # milliseconds
                    stop_time = stop_time / 1000
                scenario_end_times.append(stop_time)

    # Set global execution times from actual scenario data
    if scenario_start_times and scenario_end_times:
        global_vars._execution_start_time = min(scenario_start_times)
        global_vars._execution_end_time = max(scenario_end_times)
    else:
        # Fallback to current time if no scenario data available
        global_vars._execution_start_time = time.time() - 60
        global_vars._execution_end_time = time.time()

    try:
        # Generate HTML report (json_data already loaded above)
        print(f"Generating HTML report from {json_file_path}...")
        report_html.generate_report(json_data)

        html_path = os.path.join(output_dir, 'report.html')
        print(f"HTML report generated successfully: {html_path}")

        return html_path

    except FileNotFoundError:
        print(f"Error: JSON file not found: {json_file_path}")
        return None
    except json.JSONDecodeError as e:
        print(f"Error: Invalid JSON format in {json_file_path}: {e}")
        return None
    except Exception as e:
        print(f"Error generating HTML report: {e}")
        import traceback
        traceback.print_exc()
        return None


def main():
    """Main entry point for command line usage."""
    if len(sys.argv) < 2:
        print("Usage: python generate_html_from_json.py <path_to_report.json> [output_directory]")
        print("\nExamples:")
        print("  python generate_html_from_json.py output/report.json")
        print("  python generate_html_from_json.py output/report.json my_reports/")
        sys.exit(1)

    json_file = sys.argv[1]
    output_dir = sys.argv[2] if len(sys.argv) > 2 else None

    if not os.path.exists(json_file):
        print(f"Error: File not found: {json_file}")
        sys.exit(1)

    result = generate_html_from_json(json_file, output_dir)
    sys.exit(0 if result else 1)


if __name__ == "__main__":
    main()



import logging

from behave import given, then


@given('image attachment functionality is required')
def step_require_image_attachment(context):
    """Mark image attachment functionality as required for this scenario."""
    context.require_image_attachments = True


@then('image attachment dependencies should be available')
def step_validate_image_dependencies(context):
    """Validate that image attachment dependencies are available when required."""
    if hasattr(context, 'require_image_attachments') and context.require_image_attachments:
        try:
            from behavex_images import image_attachments
            logging.info("Image attachment dependencies are available")
        except (ImportError, ModuleNotFoundError) as e:
            raise AssertionError(f"Required dependency 'behavex-images' is not available: {e}")
    else:
        logging.info("Image attachment dependencies not required for this test")



# -*- coding: utf-8 -*-

"""BehaveX - Agile test wrapper on top of Behave (BDD).

This module provides the main entry point for running BehaveX tests,
including setup, execution, and reporting.
"""

# pylint: disable=W0703

# __future__ has been added to maintain compatibility
from __future__ import absolute_import, print_function

# Standard library imports
import codecs
import concurrent  # pyright: ignore[reportUnusedImport]
import copy
import json
import logging
import logging.config  # pyright: ignore[reportUnusedImport]
import multiprocessing
import os
import os.path
import platform
import signal
import sys
import time
import traceback
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures.process import BrokenProcessPool
from multiprocessing import active_children
from multiprocessing.managers import DictProxy
from tempfile import gettempdir
from typing import Any, Dict

# Third-party imports
from behave.configuration import Configuration
from behave.model import Feature, Scenario, ScenarioOutline
from behave.runner import Runner

# Local imports
# noinspection PyUnresolvedReferences
from behavex import conf_mgr
from behavex.arguments import BEHAVE_ARGS, BEHAVEX_ARGS, parse_arguments
from behavex.conf_mgr import ConfigRun, get_env, get_param
from behavex.environment import extend_behave_hooks
from behavex.execution_singleton import ExecutionSingleton
from behavex.global_vars import global_vars
from behavex.outputs import report_xml
from behavex.outputs.formatter_manager import (DEFAULT_FORMATTER_DIR,
                                               FormatterManager)
from behavex.outputs.report_json import (generate_execution_info,
                                         get_environment_details)
from behavex.outputs.report_utils import (get_overall_status,
                                          match_for_execution,
                                          pretty_print_time,
                                          retry_file_operation, text)
from behavex.progress_bar import ProgressBar
from behavex.utils import (IncludeNameMatch, IncludePathsMatch, MatchInclude,
                           cleanup_folders, configure_logging,
                           copy_bootstrap_html_generator,
                           create_execution_complete_callback_function,
                           expand_paths, explore_features, generate_reports,
                           get_feature_and_scenario_line, get_feature_order,
                           get_json_results, get_logging_level,
                           get_scenario_order, get_scenario_tags,
                           get_scenarios_instances, get_text,
                           join_feature_reports, join_scenario_reports,
                           len_scenarios, print_env_variables, print_parallel,
                           set_behave_tags, set_env_variable,
                           set_environ_config, set_system_paths)

EXIT_OK = 0
EXIT_ERROR = 1
EXECUTION_BLOCKED_MSG = (
    'Some of the folders or files are being used by another '
    'program. Please, close them and try again...'
)

os.environ.setdefault('EXECUTION_CODE', '1')
match_include = None
include_path_match = None
include_name_match = None


def main():
    """BehaveX starting point.

    Parses command-line arguments and initiates the test run.
    """
    args = sys.argv[1:]
    exit_code = 1  # Default to error
    try:
        exit_code = run(args)
    except SystemExit as e:
        # Handle cases where scenarios call exit() directly
        if hasattr(e, 'code') and e.code is not None:
            exit_code = e.code
        else:
            exit_code = 1
        # Ensure we print the exit code even when scenarios crash
        print('Exit code: {}'.format(exit_code))
    except Exception as e:
        # Handle unexpected exceptions
        print(f'Unexpected error: {e}')
        exit_code = 1
        print('Exit code: {}'.format(exit_code))

    try:
        exit(exit_code)
    except:
        # force exit
        sys.exit(exit_code)

def run(args):
    """Run BehaveX with the given arguments.

    Args:
        args (list): Command-line arguments.

    Returns:
        int: Exit code indicating success or failure.
    """
    global match_include
    global include_path_match
    global include_name_match
    args_parsed = parse_arguments(args)
    set_environ_config(args_parsed)
    ConfigRun().set_args(args_parsed)
    execution_code, rerun_list = setup_running_failures(args_parsed)
    if rerun_list:
        paths = ",".join(rerun_list).replace('\n', '')
        os.environ['FEATURES_PATH'] = paths
        global_vars.rerun_failures = True
        if execution_code == EXIT_ERROR:
            return EXIT_ERROR
    else:
        # Handle paths parameter
        if len(get_param('paths')) > 0:
            paths = ",".join(expand_paths(get_param('paths'))).replace('\n', '')
            os.environ['FEATURES_PATH'] = paths

        # Handle include_paths parameter
        if len(get_param('include_paths')) > 0:
            include_paths = ",".join(expand_paths(get_param('include_paths'))).replace('\n', '')
            features_path = os.environ.get('FEATURES_PATH', '')
            if features_path == '' or not os.path.exists(features_path):
                os.environ['FEATURES_PATH'] = include_paths
            else:
                os.environ['FEATURES_PATH'] = features_path + ',' + include_paths
    features_path = os.environ.get('FEATURES_PATH')
    if features_path == '' or features_path is None:
        os.environ['FEATURES_PATH'] = 'features'
    _set_env_variables(args_parsed)
    set_system_paths()
    cleanup_folders()
    copy_bootstrap_html_generator()
    configure_logging(args_parsed)
    match_include = MatchInclude()
    include_path_match = IncludePathsMatch()
    include_name_match = IncludeNameMatch()

    return launch_behavex()


def setup_running_failures(args_parsed):
    """Setup the environment for rerunning failed tests.

    Args:
        args_parsed (Namespace): Parsed command-line arguments.

    Returns:
        tuple: Exit code and list of paths to rerun.
    """
    failures_path = args_parsed.rerun_failures
    if failures_path:
        failures_path = os.path.normpath(failures_path)
        if not os.path.isabs(failures_path):
            failures_path = os.path.abspath(failures_path)
        if os.path.exists(failures_path):
            set_env_variable('RERUN_FAILURES', args_parsed.rerun_failures)
            with open(failures_path, 'r') as failures_file:
                content = failures_file.read()
                if not content:
                    print('\nThere are no failing test scenarios to run.')
                    return EXIT_ERROR, None
                return EXIT_OK, content.split(",")
        else:
            print('\nThe specified failing scenarios filename was not found: {}'.format(failures_path))
            return EXIT_ERROR, None
    else:
        return EXIT_OK, None


def init_multiprocessing(idQueue, parallel_delay):
    try:
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        # Retrieve one of the unique IDs
        worker_id = idQueue.get()
        # Use the unique ID to name the process
        multiprocessing.current_process().name = f'behave_worker-{worker_id}'
        # Add an initial delay to avoid all processes starting at the same time
        if isinstance(parallel_delay, int) and parallel_delay > 0:
            time.sleep(parallel_delay * worker_id / 1000.0)
    except Exception as e:
        logging.error(f"Exception in init_multiprocessing: {e}")
        raise


def launch_behavex():
    """Launch the BehaveX test execution in the specified parallel mode.

    Returns:
        int: Exit code indicating success or failure.
    """
    json_reports = []
    execution_codes = []
    results = None
    config = conf_mgr.get_config()
    features_path = os.environ.get('FEATURES_PATH', '')
    parallel_scheme = get_param('parallel_scheme')
    if get_param('dry_run'):
        parallel_processes = 1
        show_progress_bar = False
    else:
        parallel_processes = get_param('parallel_processes')
        parallel_processes = 1 if parallel_processes <= 1 else parallel_processes
        show_progress_bar = get_param('show_progress_bar')
    multiprocess = (
        True
        if parallel_processes > 1 and not get_param('dry_run')
        else False
    )
    set_behave_tags()
    scenario = False
    notify_missing_features(features_path)
    features_list = {}
    for path in features_path.split(','):
        features_list[path] = explore_features(path)
    updated_features_list = create_scenario_line_references(features_list)
    parallel_scheme = '' if not multiprocess else parallel_scheme
    manager = multiprocessing.Manager()
    # shared variable to track scenarios that should be run but seems to be removed from execution (using scenarios.remove)
    shared_removed_scenarios = manager.dict()
    lock = manager.Lock()
    # Create a queue containing unique IDs from 0 to the number of parallel processes - 1
    # These IDs will be attributed to the process when they will be initialized
    idQueue = manager.Queue()
    for i in range(parallel_processes):
        idQueue.put(i)
    parallel_delay = get_param('parallel_delay')
    process_pool = ProcessPoolExecutor(max_workers=parallel_processes,
                                       initializer=init_multiprocessing,
                                       initargs=(idQueue, parallel_delay))
    global_vars.execution_start_time = time.time()
    totals = {"features": {"passed": 0, "failed": 0, "error": 0, "skipped": 0, "untested": 0},
              "scenarios": {"passed": 0, "failed": 0, "error": 0, "skipped": 0, "untested": 0}}
    failures = []  # Initialize before try block to ensure it's always defined
    try:
        config = ConfigRun()
        if parallel_processes == 1 or get_param('dry_run'):
            # Executing without parallel processes
            if get_param('dry_run'):
                print('Obtaining information about the reporting scope...')
            if global_vars.rerun_failures:
                all_paths = features_path.split(",")
            else:
                all_paths = [key for key in updated_features_list]
            if len(all_paths) > 0:
                execution_codes, json_reports = execute_tests(features_path=all_paths,
                                                                feature_filename=None,
                                                                feature_json_skeleton=None,
                                                                scenarios_to_run_in_feature=None,
                                                                scenario_line=None,
                                                                multiprocess=False,
                                                                config=config,
                                                                lock=None,
                                                                shared_removed_scenarios=None)
            else:
                execution_codes, json_reports = (0, [{'environment': [], 'features': [], 'steps_definition': []}])
        elif parallel_scheme == 'scenario':
            execution_codes, json_reports = launch_by_scenario(updated_features_list,
                                                            process_pool,
                                                            lock,
                                                            shared_removed_scenarios,
                                                            show_progress_bar)
            scenario = True
        elif parallel_scheme == 'feature':
            execution_codes, json_reports = launch_by_feature(updated_features_list,
                                                            process_pool,
                                                            lock,
                                                            show_progress_bar)
        wrap_up_process_pools(process_pool, json_reports, scenario)

        if get_param('dry_run'):
            print_parallel('execution.dry_run.completed', get_env('OUTPUT'))

        remove_temporary_files(parallel_processes, json_reports)

        failing_non_muted_tests = False
        # TODO: Replace logs below with test execution logs when an unexpected error occurs
        # behave_log_file = os.path.join(output_folder, 'behavex', 'logs', str(scenario['id_feature']), 'behave.log')
        # behave_log_file = os.path.join(output_folder, 'behavex', 'logs', str(json_test_configuration['id']), 'behave.log')
        results = get_json_results()
        processed_feature_filenames = []
        if results:
            for feature in results['features']:
                processed_feature_filenames.append(feature['filename'])
                filename = feature['filename']
                if feature['status'] == 'failed':
                    totals['features']['failed'] += 1
                elif feature['status'] == 'error' or feature['status'] == 'undefined':
                    totals['features']['error'] += 1
                elif feature['status'] == 'passed':
                    totals['features']['passed'] += 1
                elif feature['status'] == 'untested':
                    totals['features']['untested'] += 1
                else:
                    totals['features']['skipped'] += 1
                    totals['scenarios']['skipped'] += len(feature['scenarios'])
                    continue
                for scenario in feature['scenarios']:
                    if scenario['status'] == 'failed':
                        totals['scenarios']['failed'] += 1
                        failures.append('{}:{}'.format(filename, scenario['line']))
                        if 'MUTE' not in scenario['tags']:
                            failing_non_muted_tests = True
                    elif scenario['status'] == 'error' or scenario['status'] == 'undefined':
                        totals['scenarios']['failed'] += 1
                        failures.append('{}:{}'.format(filename, scenario['line']))
                        if 'MUTE' not in scenario['tags']:
                            failing_non_muted_tests = True
                    elif scenario['status'] == 'passed':
                        totals['scenarios']['passed'] += 1
                    elif scenario['status'] == 'untested':
                        totals['scenarios']['untested'] += 1
                    else:
                        totals['scenarios']['skipped'] += 1
            if failures:
                failures_file_path = os.path.join(get_env('OUTPUT'), global_vars.report_filenames['report_failures'])
                with open(failures_file_path, 'w') as failures_file:
                    failures_file.write(','.join(failures))
        # Calculates final exit code. execution_codes is 1 only if an execution exception arises
        if isinstance(execution_codes, list):
            execution_failed = True if sum(execution_codes) > 0 else False
            execution_interrupted_or_crashed = True if any([code == 2 for code in execution_codes]) else False
        else:
            execution_failed = True if execution_codes > 0 else False
            execution_interrupted_or_crashed = True if execution_codes == 2 else False
        exit_code = (EXIT_ERROR if (execution_failed and failing_non_muted_tests) or execution_interrupted_or_crashed else EXIT_OK)
    except KeyboardInterrupt as ex:
        print('Caught KeyboardInterrupt, terminating workers')
        try:
            while active_children():
                for process in active_children():  # Terminate all active child processes
                    process.terminate()  # Forcefully terminate each process
                time.sleep(1)
            process_pool.shutdown(wait=False)
        except Exception as e:
            print(f"Error during shutdown: {e}")
        exit_code = EXIT_ERROR
    if multiprocess:
        print_execution_summary(totals, failures, results)  # failures initialized above
    if results and results['features'] and not get_param('formatter'):
        print('\nHTML output report is located at: {}'.format(os.path.join(get_env('OUTPUT'), "report.html")))
    print('Exit code: {}'.format(exit_code))
    return exit_code


def print_execution_summary(totals, failures, results):
    """Print execution summary including failing scenarios and count totals.

    Args:
        totals (dict): Dictionary containing execution totals by type
        failures (list): List of failing scenario identifiers
        results (dict): Full results dictionary with feature/scenario details
    """
    untested_features = totals['features']['untested']
    untested_scenarios = totals['scenarios']['untested']
    error_features = totals['features']['error']
    error_scenarios = totals['scenarios']['error']

    untested_features_msg = ', {} untested'.format(untested_features) if untested_features > 0 else ''
    untested_scenarios_msg = ', {} untested'.format(untested_scenarios) if untested_scenarios > 0 else ''
    error_features_msg = ', {} error'.format(error_features) if error_features > 0 else ''
    error_scenarios_msg = ', {} error'.format(error_scenarios) if error_scenarios > 0 else ''

    # Print failing scenarios before summary (same format as behave single execution)
    if failures and results:
        # Collect failed and errored scenarios with their details
        failed_scenarios = []
        errored_scenarios = []

        # Group failures by status type
        for feature in results['features']:
            filename = feature['filename']
            for scenario in feature['scenarios']:
                scenario_key = f"{filename}:{scenario['line']}"
                if scenario_key in failures:
                    scenario_line = f"  {filename}:{scenario['line']}  {scenario['name']}"
                    if scenario['status'] == 'failed':
                        failed_scenarios.append(scenario_line)
                    elif scenario['status'] in ['error', 'undefined']:
                        errored_scenarios.append(scenario_line)

        # Print errored scenarios first (if any)
        if errored_scenarios:
            print("\nErrored scenarios:")
            for scenario_line in errored_scenarios:
                print(scenario_line)

        # Print failed scenarios (if any)
        if failed_scenarios:
            print("\nFailing scenarios:")
            for scenario_line in failed_scenarios:
                print(scenario_line)

        print()  # Empty line before summary

    def plural_char(n): return 's' if n != 1 else ''
    print('\n{} feature{} passed, {} failed{}, {} skipped{}'.format(totals['features']['passed'],
                                                                      plural_char(totals['features']['passed']),
                                                                      totals['features']['failed'],
                                                                      error_features_msg,
                                                                      totals['features']['skipped'],
                                                                      untested_features_msg))
    print('{} scenario{} passed, {} failed{}, {} skipped{}'.format(totals['scenarios']['passed'],
                                                                    plural_char(totals['scenarios']['passed']),
                                                                    totals['scenarios']['failed'],
                                                                    error_scenarios_msg,
                                                                    totals['scenarios']['skipped'],
                                                                    untested_scenarios_msg))
    # Calculate steps summary
    steps_totals = {"passed": 0, "failed": 0, "error": 0, "skipped": 0, "untested": 0, "undefined": 0}
    if results and results['features']:
        for feature in results['features']:
            for scenario in feature['scenarios']:
                for step in scenario.get('steps', []):
                    step_status = step.get('status', 'skipped')
                    if step_status == 'undefined':
                        steps_totals['undefined'] += 1
                    elif step_status in steps_totals:
                        steps_totals[step_status] += 1
                    else:
                        steps_totals['skipped'] += 1

    untested_steps_msg = ', {} untested'.format(steps_totals['untested']) if steps_totals['untested'] > 0 else ''
    undefined_steps_msg = ', {} undefined'.format(steps_totals['undefined']) if steps_totals['undefined'] > 0 else ''
    error_steps_msg = ', {} error'.format(steps_totals['error']) if steps_totals['error'] > 0 else ''
    print('{} steps passed, {} failed{}, {} skipped{}{}'.format(steps_totals['passed'],
                                                                steps_totals['failed'],
                                                                error_steps_msg,
                                                                steps_totals['skipped'],
                                                                untested_steps_msg,
                                                                undefined_steps_msg))
    print('Took: {}'.format(pretty_print_time(global_vars.execution_end_time - global_vars.execution_start_time)))


def notify_missing_features(features_path):
    """Notify if any features are missing in the specified path.

    Args:
        features_path (str): Path to the features.
    """
    if global_vars.rerun_failures:
        all_paths = features_path.split(",")
        for path in all_paths:
            include_path = path.partition(':')[0]
            if not os.path.exists(os.path.normpath(include_path)):
                print_parallel('path.not_found', os.path.realpath(include_path))


def create_scenario_line_references(features):
    """Create references for scenario lines in the features.

    Args:
        features (dict): Dictionary of features and their scenarios.

    Returns:
        dict: Updated features with scenario line references.
    """
    updated_features = {}
    for feature_path, scenarios in features.items():
        for scenario in scenarios:
            if global_vars.rerun_failures or ".feature:" in feature_path:
                if feature_path not in updated_features:
                    updated_features[feature_path] = []
                if isinstance(scenario, ScenarioOutline):
                    for scenario_outline_instance in scenario.scenarios:
                        outline_scenario_line = get_feature_and_scenario_line(scenario_outline_instance.name)[1]
                        if outline_scenario_line and scenario_outline_instance.line == int(outline_scenario_line):
                            if scenario_outline_instance not in updated_features[feature_path]:
                                updated_features[feature_path].append(scenario_outline_instance)
                            break
                else:
                    scenario_line = get_feature_and_scenario_line(feature_path)[1]
                    if scenario_line and scenario.line == int(scenario_line):
                        if scenario not in updated_features[feature_path]:
                            updated_features[feature_path].append(scenario)
            else:
                updated_features_path = scenario.feature.filename
                if updated_features_path not in updated_features:
                    updated_features[updated_features_path] = []
                if isinstance(scenario, ScenarioOutline):
                    for scenario_outline_instance in scenario.scenarios:
                        if scenario_outline_instance not in updated_features[updated_features_path]:
                            updated_features[updated_features_path].append(scenario_outline_instance)
                else:
                    if scenario not in updated_features[updated_features_path]:
                        updated_features[updated_features_path].append(scenario)
    return updated_features


def _wait_for_futures(futures, execution_codes, json_reports):
    """Helper function to wait for futures and handle exceptions"""
    for future in futures:
        try:
            future.result()
        except (KeyboardInterrupt, SystemExit):
            # Allow KeyboardInterrupt (Ctrl+C) and SystemExit to propagate to top level
            raise
        except:
            e = sys.exc_info()[1]
            if isinstance(e, BrokenProcessPool):
                print_parallel('process.pool.broken.main', str(e))
                # Mark as failed execution but continue processing remaining futures
                execution_codes.append(1)
            else:
                print_parallel('parallel.process.error', str(e))
                execution_codes.append(1)


def launch_by_feature(features,
                      process_pool,
                      lock,
                      show_progress_bar):
    """Launch tests by feature in parallel.

    Args:
        features (dict): Dictionary of features and their scenarios.
        process_pool (ProcessPoolExecutor): Process pool executor.
        lock (Lock): Multiprocessing lock.
        show_progress_bar (bool): Whether to show the progress bar.

    Returns:
        tuple: Execution codes and JSON reports.
    """
    json_reports = []
    execution_codes = []

    # Cache the order_tests parameter to avoid multiple get_param calls
    order_tests_enabled = get_param('order_tests')
    order_tests_strict = get_param('order_tests_strict')
    # If order_tests_strict is enabled, automatically enable order_tests
    order_tests_enabled = order_tests_enabled or order_tests_strict
    order_tag_prefix = get_param('order_tag_prefix') if order_tests_enabled else None

    serial_features = []
    parallel_features = []
    for features_path in features:
        feature = features[features_path][0].feature
        pure_feature_path, scenario_line = get_feature_and_scenario_line(features_path)
        feature_filename = feature.filename if not scenario_line else "{}:{}".format(pure_feature_path, scenario_line)

        feature_info = {"feature_filename": feature_filename,
                       "feature_json_skeleton": _get_feature_json_skeleton(feature)}

        # Only calculate feature order if ordering is enabled
        if order_tests_enabled:
            # For feature-level execution, use ORDER tags from the feature itself
            feature_info["feature_order"] = get_feature_order(feature, order_tag_prefix)

        if 'SERIAL' in feature.tags:
            serial_features.append(feature_info)
        else:
            parallel_features.append(feature_info)
    # Sort features by execution order if enabled (only the categorized features that will actually run)
    if order_tests_enabled:
        serial_features.sort(key=lambda f: f.get("feature_order", 9999))
        parallel_features.sort(key=lambda f: f.get("feature_order", 9999))

    if show_progress_bar:
        total_features = len(serial_features) + len(parallel_features)
        global_vars.progress_bar_instance = _get_progress_bar_instance(parallel_scheme="feature",
                                                                       total_elements=total_features)
        if global_vars.progress_bar_instance:
            global_vars.progress_bar_instance.start()
    if serial_features:
        print_parallel('feature.serial_execution')
        for serial_feature in serial_features:
            execution_code, map_json = execute_tests(features_path=None,
                                                     feature_filename=serial_feature["feature_filename"],
                                                     feature_json_skeleton=serial_feature["feature_json_skeleton"],
                                                     scenarios_to_run_in_feature=None,
                                                     scenario_line=None,
                                                     multiprocess=True,
                                                     config=ConfigRun(),
                                                     lock=None,
                                                     shared_removed_scenarios=None)
            json_reports += [map_json]
            execution_codes.append(execution_code)
            if global_vars.progress_bar_instance:
                global_vars.progress_bar_instance.update()
    print_parallel('feature.running_parallels')
    parallel_processes = []


    # Prepare features grouped by order
    if order_tests_strict:
        # Each feature gets its own group for strict sequential ordering
        features_by_order = {}
        # Sort features by order first
        sorted_features = sorted(parallel_features, key=lambda f: f.get("feature_order", 9999))
        for index, parallel_feature in enumerate(sorted_features):
            features_by_order[index] = [parallel_feature]
    elif order_tests_enabled:
        # Sort features by order but allow parallel execution
        parallel_features.sort(key=lambda f: f.get("feature_order", 9999))
        features_by_order = {9999: parallel_features}
    else:
        # Put all features in a single group for original behavior
        features_by_order = {9999: parallel_features}

    # Execute features by order groups
    for order in sorted(features_by_order.keys()):
        feature_group = features_by_order[order]
        group_futures = []

        # Submit all features of the current order
        for parallel_feature in feature_group:
            feature_filename = parallel_feature["feature_filename"]
            feature_json_skeleton = parallel_feature["feature_json_skeleton"]
            future = process_pool.submit(execute_tests,
                                         features_path=None,
                                         feature_filename=feature_filename,
                                         feature_json_skeleton=feature_json_skeleton,
                                         scenarios_to_run_in_feature=None,
                                         scenario_line=None,
                                         multiprocess=True,
                                         config=ConfigRun(),
                                         lock=lock,
                                         shared_removed_scenarios=None)
            parallel_processes.append(future)
            future.add_done_callback(create_execution_complete_callback_function(
                execution_codes,
                json_reports,
                global_vars.progress_bar_instance,
            ))
            group_futures.append(future)

        # Wait for completion of this order group before proceeding to the next
        _wait_for_futures(group_futures, execution_codes, json_reports)

    parallel_processes.clear()
    return execution_codes, json_reports


def launch_by_scenario(features,
                       process_pool,
                       lock,
                       shared_removed_scenarios,
                       show_progress_bar):
    """Launch tests by scenario in parallel.

    Args:
        features (dict): Dictionary of features and their scenarios.
        process_pool (ProcessPoolExecutor): Process pool executor.
        lock (Lock): Multiprocessing lock.
        shared_removed_scenarios (dict): Shared dictionary of removed scenarios.
        show_progress_bar (bool): Whether to show the progress bar.

    Returns:
        tuple: Execution codes and JSON reports.
    """
    json_reports = []
    execution_codes = []
    parallel_scenarios = []
    serial_scenarios = []
    duplicated_scenarios = {}
    total_scenarios_to_run = {}
    features_with_no_scen_desc = []
    parallel_processes = []

    # Cache the order_tests parameter to avoid multiple get_param calls
    order_tests_enabled = get_param('order_tests')
    order_tests_strict = get_param('order_tests_strict')
    # If order_tests_strict is enabled, automatically enable order_tests
    order_tests_enabled = order_tests_enabled or order_tests_strict
    order_tag_prefix = get_param('order_tag_prefix') if order_tests_enabled else None
    for features_path, scenarios in features.items():
        scenarios_instances = get_scenarios_instances(scenarios)

        for scenario in scenarios_instances:
            if include_path_match(scenario.filename, scenario.line) \
                    and include_name_match(scenario.name):
                scenario_tags = get_scenario_tags(scenario)
                if match_for_execution(scenario_tags):
                    if scenario.name == "":
                        features_with_no_scen_desc.append(scenario.filename)
                    feature_json_skeleton = _get_feature_json_skeleton(scenario)
                    feature_filename = scenario.feature.filename
                    scenario_information = {"features_path": features_path,
                                            "feature_filename": feature_filename,
                                            "feature_json_skeleton": feature_json_skeleton,
                                            "scenario_line": scenario.line}

                    # Only calculate scenario order if ordering is enabled
                    if order_tests_enabled:
                        scenario_information["scenario_order"] = get_scenario_order(scenario, order_tag_prefix)
                    total_scenarios_to_run[feature_filename] = total_scenarios_to_run.setdefault(feature_filename, 0) + 1
                    if 'SERIAL' in scenario_tags:
                        if scenario_information in serial_scenarios:
                            duplicated_scenarios.setdefault(scenario.filename, []).append(scenario.name)
                        serial_scenarios.append(scenario_information)
                    else:
                        if scenario_information in parallel_scenarios:
                            duplicated_scenarios.setdefault(scenario.filename, []).append(scenario.name)
                        parallel_scenarios.append(scenario_information)
    if show_progress_bar:
        global_vars.progress_bar_instance = _get_progress_bar_instance(parallel_scheme="scenario",
                                                                       total_elements=sum(total_scenarios_to_run.values()))
        if global_vars.progress_bar_instance:
            global_vars.progress_bar_instance.start()
    if duplicated_scenarios:
        print_parallel('scenario.duplicated_scenarios',
                       json.dumps(duplicated_scenarios, indent=4))
        exit(1)
    if features_with_no_scen_desc:
        print_parallel('feature.empty_scenario_descriptions',
                       '\n* '.join(features_with_no_scen_desc))
        exit(1)

    # Sort scenarios by execution order if enabled (only the filtered scenarios that will actually run)
    if order_tests_enabled:
        serial_scenarios.sort(key=lambda s: s.get("scenario_order", 9999))
        parallel_scenarios.sort(key=lambda s: s.get("scenario_order", 9999))

    if serial_scenarios:
        print_parallel('scenario.serial_execution')
        for scen_info in serial_scenarios:
            scenarios_to_run_in_feature = total_scenarios_to_run[scen_info["feature_filename"]]
            execution_code, json_report = execute_tests(features_path=scen_info["features_path"],
                                                        feature_filename=scen_info["feature_filename"],
                                                        feature_json_skeleton=scen_info["feature_json_skeleton"],
                                                        scenarios_to_run_in_feature=scenarios_to_run_in_feature,
                                                        scenario_line=scen_info["scenario_line"],
                                                        multiprocess=True,
                                                        config=ConfigRun(),
                                                        lock=None,
                                                        shared_removed_scenarios=shared_removed_scenarios)
            execution_codes.append(execution_code)
            json_reports.append(json_report)
            if global_vars.progress_bar_instance:
                global_vars.progress_bar_instance.update()
    if parallel_scenarios:
        print_parallel('scenario.running_parallels')


        # Prepare scenarios grouped by order
        if order_tests_strict:
            # Each scenario gets its own group for strict sequential ordering
            scenarios_by_order = {}
            # Sort scenarios by order first
            sorted_scenarios = sorted(parallel_scenarios, key=lambda s: s.get("scenario_order", 9999))
            for index, scenario_info in enumerate(sorted_scenarios):
                scenarios_by_order[index] = [scenario_info]
        elif order_tests_enabled:
            # Sort scenarios by order but allow parallel execution
            parallel_scenarios.sort(key=lambda s: s.get("scenario_order", 9999))
            scenarios_by_order = {9999: parallel_scenarios}
        else:
            # Put all scenarios in a single group for original behavior
            scenarios_by_order = {9999: parallel_scenarios}

        # Execute scenarios by order groups
        for order in sorted(scenarios_by_order.keys()):
            scenario_group = scenarios_by_order[order]
            group_futures = []

            # Submit all scenarios of the current order
            for scenario_information in scenario_group:
                scenarios_to_run_in_feature = total_scenarios_to_run[scenario_information["feature_filename"]]
                features_path = scenario_information["features_path"]
                feature_filename = scenario_information["feature_filename"]
                feature_json_skeleton = scenario_information["feature_json_skeleton"]
                scenario_line = scenario_information["scenario_line"]
                future = process_pool.submit(execute_tests,
                                            features_path=features_path,
                                            feature_filename=feature_filename,
                                            feature_json_skeleton=feature_json_skeleton,
                                            scenarios_to_run_in_feature=scenarios_to_run_in_feature,
                                            scenario_line=scenario_line,
                                            multiprocess=True,
                                            config=ConfigRun(),
                                            lock=lock,
                                            shared_removed_scenarios=shared_removed_scenarios
                                            )
                parallel_processes.append(future)
                future.add_done_callback(create_execution_complete_callback_function(
                    execution_codes,
                    json_reports,
                    global_vars.progress_bar_instance
                ))
                group_futures.append(future)

            # Wait for completion of this order group before proceeding to the next
            _wait_for_futures(group_futures, execution_codes, json_reports)

        parallel_processes.clear()
    return execution_codes, json_reports


def execute_tests(
        features_path,
        feature_filename,
        feature_json_skeleton,
        scenarios_to_run_in_feature,
        scenario_line,
        multiprocess,
        config,
        lock,
        shared_removed_scenarios):
    """
    Execute tests for the given feature or scenario.

    Args:
        features_path (str): Path to the features.
        feature_filename (str): Name of the feature file.
        feature_json_skeleton (str): JSON skeleton of the feature.
        scenarios_to_run_in_feature (int): Number of scenarios to run in the feature.
        scenario_line (int): Line of the scenario.
        multiprocess (bool): Whether to use multiprocessing.
        config (ConfigRun): Configuration object.
        lock (Lock): Multiprocessing lock.
        shared_removed_scenarios (dict): Shared dictionary of removed scenarios.

    Returns:
        tuple: Execution code and JSON report.
    """
    try:
        behave_args = None
        if multiprocess:
            ExecutionSingleton._instances[ConfigRun] = config
        extend_behave_hooks()
        try:
            # Execution ID is only important for multiprocessing so that
            # we can influence where output files end up
            execution_id = json.loads(feature_json_skeleton or '{}').get('id')
            behave_args = _set_behave_arguments(features_path=features_path,
                                                multiprocess=multiprocess,
                                                execution_id=execution_id,
                                                feature=feature_filename,
                                                scenario_line=scenario_line,
                                                config=config)
        except Exception as exception:
            traceback.print_exc()
            print(exception)
        execution_code, generate_report, json_results_str = _launch_behave(behave_args)
        # print("pipenv run behave {} --> Execution Code: {} --> Generate Report: {}".format(" ".join(behave_args), execution_code, generate_report))
        if generate_report:
            # print execution code
            if execution_code == 2:
                # For crashed executions, override with skeleton data if available
                if feature_json_skeleton:
                    json_output = {'environment': [],
                                   'features': [json.loads(feature_json_skeleton)],
                                   'steps_definition': []}
                    for skeleton_feature in json_output["features"]:
                        if scenario_line:
                            for skeleton_scenario in skeleton_feature["scenarios"]:
                                if str(skeleton_scenario['line']) == str(scenario_line):
                                    skeleton_scenario['status'] = 'failed'
                                    skeleton_scenario['error_msg'] = get_text('scenario.execution_crashed')
                        else:
                            skeleton_feature['status'] = 'failed'
                            skeleton_feature['error_msg'] = 'Execution crashed. No outputs could be generated.'
                            for skeleton_scenario in skeleton_feature["scenarios"]:
                                skeleton_scenario['status'] = 'failed'
                                skeleton_scenario['error_msg'] = get_text('feature.execution_crashed')
                else:
                    json_output = {'environment': [], 'features': [], 'steps_definition': []}
            else:
                # Parse JSON string from _launch_behave (disk-free approach)
                try:
                    json_output = json.loads(json_results_str)
                except (json.JSONDecodeError, ValueError) as e:
                    logging.error(f"Failed to parse JSON results from _launch_behave: {e}")
                    logging.error(f"Raw JSON string: {json_results_str}")
                    # Fallback to empty structure
                    json_output = {'environment': [], 'features': [], 'steps_definition': []}
            if scenario_line:
                json_output['features'] = filter_feature_executed(json_output,
                                                                  text(feature_filename),
                                                                  scenario_line=scenario_line)
                if len(json_output['features']) == 0 or len(json_output['features'][0]['scenarios']) == 0:
                    # Adding scenario data if the test was removed from the execution (setting it as "Untested")
                    json_output['features'] = [json.loads(feature_json_skeleton)]
                try:
                    processing_xml_feature(json_output=json_output,
                                           scenario_line=scenario_line,
                                           feature_filename=feature_filename,
                                           scenarios_to_run_in_feature=scenarios_to_run_in_feature,
                                           lock=lock,
                                           shared_removed_scenarios=shared_removed_scenarios)
                except Exception as ex:
                    logging.exception("There was a problem processing the xml file: {}".format(ex))
        else:
            json_output = {'environment': [], 'features': [], 'steps_definition': []}
        return execution_code, join_feature_reports(json_output)
    except Exception as e:
        logging.error(f"Exception in execute_tests: {e}")
        raise


def filter_feature_executed(json_output, filename, scenario_line):
    """
    Filter the executed feature from the JSON output.

    Args:
        json_output (dict): JSON output of the test execution.
        filename (str): Name of the feature file.
        scenario_line (str): Line of the scenario.
    """
    for feature in json_output.get('features', '')[:]:
        if feature.get('filename', '') == filename:
            mapping_scenarios = []
            for scenario in feature['scenarios']:
                if str(scenario['line']) == str(scenario_line):
                    mapping_scenarios.append(scenario)
            feature['scenarios'] = mapping_scenarios
            return [feature]
    return []


def _calculate_execution_code_from_runner(runner):
    """
    Calculate execution code based on Runner properties.

    This uses the efficient runner.context.failed boolean and runner.hook_failures
    instead of iterating through features/scenarios or parsing output text.

    Args:
        runner: The Runner instance after execution

    Returns:
        int: Execution code (0=success, 1=test failures, 2=errors)
    """
    try:
        # Special handling for dry runs - they should always return 0 unless aborted
        if get_param('dry_run'):
            # Only return error if explicitly aborted, otherwise dry runs are always successful
            if hasattr(runner, 'aborted') and runner.aborted:
                return 2  # Aborted execution
            return 0  # Dry runs are successful by definition (they don't actually run tests)

        # Check if runner was aborted first
        if hasattr(runner, 'aborted') and runner.aborted:
            return 2  # Aborted execution

        # Check for hook failures using runner.hook_failures (more reliable than text parsing)
        if hasattr(runner, 'hook_failures') and runner.hook_failures > 0:
            return 2  # Hook failures

        # Use context.failed boolean to determine if any tests failed
        if hasattr(runner, 'context') and runner.context:
            if hasattr(runner.context, 'failed'):
                if runner.context.failed:
                    return 1  # Test failures
                else:
                    return 0  # Success - all tests passed

        # Fallback if context or failed attribute not available
        return 0  # Assume success

    except Exception as ex:

        logging.warning(f"Error calculating execution code from runner: {ex}")
        return 1  # Default to failure on error


def _launch_behave(behave_args):
    """
    Launch Behave with the given arguments.

    Args:
        behave_args (list): List of arguments for Behave.

    Returns:
        tuple: Execution code, whether to generate a report, and JSON string of execution results.
    """
    generate_report = True
    execution_code = 0
    json_results_str = '{"environment": [], "features": [], "steps_definition": {}}'

    try:
        if behave_args is None:
            # Handle case where behave_args is None (e.g., argument parsing failed)
            logging.error("behave_args is None - argument parsing may have failed")
            execution_code = 2
            generate_report = True
            return execution_code, generate_report, json_results_str

        # Note: stdout_file logic removed since we get execution results directly from runner

        # Run behave using Runner class instead of behave_script.main()
        try:
            # Create configuration from command-line arguments
            config = Configuration(command_args=behave_args)

            # Ensure format is set (fallback to pretty if not specified via arguments)
            if not config.format:
                config.format = ['pretty']

            # Create runner instance
            runner = Runner(config)

            # Run the tests (Behave output suppressed via format configuration)
            runner.run()

            # Calculate execution code using runner internal state
            execution_code = _calculate_execution_code_from_runner(runner)

            # Extract results directly from runner and serialize to JSON string
            try:
                # Check if runner has features and they contain data
                if runner and hasattr(runner, 'features') and runner.features:
                    feature_list = generate_execution_info(runner.features)
                    json_results = {
                        'environment': get_environment_details(),
                        'features': feature_list,
                        'steps_definition': global_vars.steps_definitions,
                    }
                    json_results_str = json.dumps(json_results)
                else:
                    # Fallback for cases where runner doesn't have features or features is empty
                    # This is common in behave 1.2.6 during dry runs
                    if get_param('dry_run'):
                        logging.info("Dry run mode: runner.features empty, this is expected in behave 1.2.6")
                    else:
                        logging.warning("Runner doesn't have features attribute or features is empty, using empty results")

                    json_results = {
                        'environment': get_environment_details(),
                        'features': [],
                        'steps_definition': global_vars.steps_definitions,
                    }
                    json_results_str = json.dumps(json_results)
            except Exception as json_ex:
                # Robust fallback for any JSON serialization issues
                logging.error(f"Error processing runner results: {json_ex}")
                json_results_str = json.dumps({
                    'environment': [],
                    'features': [],
                    'steps_definition': {}
                })

            # Note: stdout file existence check removed since we get execution status directly from runner
        except SystemExit as system_exit:
            # Handle SystemExit from crashing tests (e.g., exit() calls)
            # SystemExit from crashing tests should be treated as crash/interruption (code 2)
            # This ensures execution_interrupted_or_crashed = True in final exit code calculation
            execution_code = 2
            generate_report = True

            logging.info(f'SystemExit caught from crashing test (original code: {system_exit.code}), treating as crash (code 2)')
        except Exception as runner_ex:
            # Handle runner exception
            # Log the runner exception
            logging.exception('Error using Runner class, details: ')
            logging.exception(runner_ex)

            # Set appropriate execution code
            execution_code = 1
            generate_report = True
    except KeyboardInterrupt:
        execution_code = 1
        generate_report = False
    except Exception as ex:
        execution_code = 1
        generate_report = True
        logging.exception('Unexpected error executing behave steps: ')
        logging.exception(ex)
        traceback.print_exc()
    except:
        execution_code = 2
        generate_report = True
    finally:
        # Cleanup is handled in the try/except blocks above
        pass

    # Note: stdout file merging removed since we get execution results directly from runner
    return execution_code, generate_report, json_results_str


def wrap_up_process_pools(process_pool,
                          json_reports,
                          scenario=False):
    """
    Wrap up the process pools and generate the final report.

    Args:
        process_pool (ProcessPoolExecutor): Process pool executor.
        json_reports (list): List of JSON reports.
        scenario (bool): Whether the execution was by scenario.
    """
    merged_json = None
    output = os.path.join(get_env('OUTPUT'))
    try:
        process_pool.shutdown(wait=True)
    except Exception:
        process_pool.shutdown(wait=False, cancel_futures=True)
    if type(json_reports) is list:
        if scenario:
            json_reports = join_scenario_reports(json_reports)
        merged_json = join_feature_reports(json_reports)
    else:
        merged_json = json_reports
    if global_vars.progress_bar_instance:
        global_vars.progress_bar_instance.finish()
    status_info = os.path.join(output, global_vars.report_filenames['report_overall'])
    with open(status_info, 'w') as file_info:
        over_status = {'status': get_overall_status(merged_json)}
        file_info.write(json.dumps(over_status))
    path_info = os.path.join(output, global_vars.report_filenames['report_json'])
    with open(path_info, 'w') as file_info:
        file_info.write(json.dumps(merged_json))
    generate_reports(merged_json)


def remove_temporary_files(parallel_processes, json_reports):
    """
    Remove temporary files created during the test execution.

    Args:
        parallel_processes (int): Number of parallel processes.
        json_reports (list): List of JSON reports.
    """
    path_info = os.path.join(
        os.path.join(get_env('OUTPUT'), global_vars.report_filenames['report_json'])
    )
    if os.path.exists(path_info):
        with open(path_info, 'r') as json_file:
            results_json = json.load(json_file)
            if 'features' and results_json['features']:
                return

    for i in range(parallel_processes):
        result_temp = os.path.join(gettempdir(), 'result{}.tmp'.format(i + 1))
        if os.path.exists(result_temp):
            try:
                os.remove(result_temp)
            except Exception as remove_ex:
                print(remove_ex)
        # Note: stdout file cleanup removed since we no longer generate these files

    # Note: stdout file cleanup removed since we no longer generate these files
    logger = logging.getLogger()
    logger.propagate = False
    for handler in logging.root.handlers:
        logger.removeHandler(handler)
    for handler in logger.handlers:
        logger.removeHandler(handler)
    logging._handlers = []
    console_log = logging.StreamHandler(sys.stdout)
    console_log.setLevel(get_logging_level())
    logger.addHandler(console_log)
    # removing any pending temporary files
    if type(json_reports) is not list:
        json_reports = [json_reports]
    for json_report in json_reports:
        if 'features' in json_report and json_report['features']:
            feature_name = os.path.join(get_env('OUTPUT'), u'{}.tmp'.format(json_report['features'][0]['name']))
            if os.path.exists(feature_name):
                try:
                    os.remove(feature_name)
                except Exception as remove_ex:
                    print(remove_ex)


def processing_xml_feature(json_output, scenario_line, feature_filename,
                           scenarios_to_run_in_feature=None, lock=None,
                           shared_removed_scenarios=None):
    """
    Process the XML feature and update the JSON output.

    Args:
        json_output (dict): JSON output of the test execution.
        scenario (Scenario): Scenario object.
        feature_filename (str): Name of the feature file.
        scenarios_to_run_in_feature (int): Number of scenarios to run in the feature.
        lock (Lock): Multiprocessing lock.
        shared_removed_scenarios (dict): Shared dictionary of removed scenarios.
    """
    if lock:
        lock.acquire()
    try:
        feature_contains_scenarios = True if (json_output['features'] and
                                              'scenarios' in json_output['features'][0] and
                                              json_output['features'][0]["scenarios"]) else False
        if type(shared_removed_scenarios) == DictProxy and not feature_contains_scenarios:
            # Assuming the scenario was removed from the execution (using scenarios.remove) as it is not in execution
            if feature_filename not in shared_removed_scenarios:
                shared_removed_scenarios[feature_filename] = 1
            else:
                shared_removed_scenarios[feature_filename] += 1
        if feature_contains_scenarios:
            reported_scenarios = json_output['features'][0]['scenarios']
            executed_scenarios = []
            for reported_scenario in reported_scenarios:
                if reported_scenario['line'] == scenario_line:
                    executed_scenarios.append(reported_scenario)
            json_output['features'][0]['scenarios'] = executed_scenarios
            feature_name = os.path.join(
                get_env('OUTPUT'), u'{}.tmp'.format(os.path.basename(feature_filename))
            )
            processed_feature_data = json_output['features'][0]
            processed_feature_data['scenarios'] = executed_scenarios
            if os.path.exists(feature_name):
                for _ in range(0, 10):
                    try:
                        processed_feature_data = json.load(open(feature_name, 'r'))
                        with open(feature_name, 'w') as feature_file:
                            for executed_scenario in executed_scenarios:
                                processed_feature_data['scenarios'].append(executed_scenario)
                            json.dump(processed_feature_data, feature_file)
                        break
                    except Exception as ex:
                        logging.debug(ex)
                        logging.debug('Retrying reading from {}'.format(feature_name))
                        time.sleep(1)
            else:
                with codecs.open(feature_name, 'w', 'utf8') as feature_file:
                    json.dump(processed_feature_data, feature_file)
            # calculate the total number of scenarios that should be executed
            removed_scenarios = 0
            if shared_removed_scenarios and feature_filename in shared_removed_scenarios:
                removed_scenarios = shared_removed_scenarios[feature_filename]
            if scenarios_to_run_in_feature is None:
                total_scenarios = len_scenarios(feature_filename) - removed_scenarios
            else:
                total_scenarios = scenarios_to_run_in_feature - removed_scenarios
            if len(processed_feature_data['scenarios']) == total_scenarios:
                try:
                    report_xml.export_feature_to_xml(processed_feature_data, False)
                except Exception as ex:
                    traceback.print_exc()
                    print(ex)
                finally:
                    path_tmp = u'{}.tmp'.format(feature_name[:-4])
                    os.remove(path_tmp)
    except Exception as ex:
        raise ex
    finally:
        if lock:
            lock.release()


def _set_env_variables(args):
    """
    Set environment variables based on the given arguments.

    Args:
        args (Namespace): Parsed command-line arguments.
    """
    output_folder = os.path.normpath(get_env('output'))
    if os.path.isabs(output_folder):
        set_env_variable('OUTPUT', output_folder)
    else:
        set_env_variable('OUTPUT', os.path.abspath(output_folder))
    _store_tags_to_env_variable(args.tags)
    if get_param('include_paths'):
        set_env_variable('INCLUDE_PATHS', get_param('include_paths'))
    if get_param('include'):
        set_env_variable('INCLUDE', get_param('include'))
    if get_param('name'):
        set_env_variable('NAME', args.name)
    if get_param('formatter'):
        formatter_outdir = get_param('formatter_outdir', '')
        formatter_spec = get_param('formatter')

        # If no formatter output directory is specified, use the default
        if not formatter_outdir:
            formatter_outdir = DEFAULT_FORMATTER_DIR

        # Special handling for custom formatters with default settings
        if formatter_outdir == DEFAULT_FORMATTER_DIR and formatter_spec:
            # Try to dynamically get the formatter's preferred output directory
            custom_formatter_dir = FormatterManager.get_formatter_output_dir(formatter_spec)
            if custom_formatter_dir:
                # Use the formatter's preferred directory to maintain consistency
                set_env_variable('LOGS', os.path.join(get_env('output'), custom_formatter_dir))
            else:
                # Fallback to default formatter directory
                set_env_variable('LOGS', os.path.join(get_env('output'), formatter_outdir))
        else:
            set_env_variable('LOGS', os.path.join(get_env('output'), formatter_outdir))
    else:
        set_env_variable('LOGS', os.path.join(get_env('output'), 'outputs', 'logs'))
    for arg in BEHAVEX_ARGS[4:]:
        set_env_variable(arg.upper(), get_param(arg))

    set_env_variable('TEMP', os.path.join(get_env('output'), 'temp'))
    set_env_variable('LOGGING_LEVEL', get_logging_level())
    if platform.system() == 'Windows':
        set_env_variable('HOME', os.path.abspath('.\\'))
    set_env_variable('DRY_RUN', get_param('dry_run'))
    print_env_variables(
        [
            'HOME',
            'CONFIG',
            'OUTPUT',
            'TAGS',
            'PARALLEL_SCHEME',
            'PARALLEL_PROCESSES',
            'FEATURES_PATH',
            'TEMP',
            'LOGS',
            'LOGGING_LEVEL'
        ]
    )


def _store_tags_to_env_variable(tags):
    """
    Store tags to the environment variable.

    Args:
        tags (list): List of tags.
    """
    config = conf_mgr.get_config()
    tags_skip = config['test_run']['tags_to_skip']
    if isinstance(tags_skip, str) and tags_skip:
        tags_skip = [tags_skip]
    else:
        tags_skip = tags_skip
    tags = tags if tags is not None else []
    tags_skip = [tag for tag in tags_skip if tag not in tags]
    tags = tags + ['~@{0}'.format(tag) for tag in tags_skip] if tags else []
    if tags:
        for tag in tags:
            existing_tags = get_env('TAGS')
            if existing_tags:
                set_env_variable('TAGS', existing_tags + ';' + tag)
            else:
                set_env_variable('TAGS', tag)
    else:
        set_env_variable('TAGS', '')


def _set_behave_arguments(features_path, multiprocess, execution_id=None, feature=None, scenario_line=None, config=None):
    """
    Set the arguments for Behave framework based on the given parameters.

    Args:
        features_path (str): Path to the features.
        multiprocess (bool): Whether to use multiprocessing.
        execution_id (str): Execution ID.
        feature (Feature): Feature object.
        scenario_line (int): Scenario line.
        config (ConfigRun): Configuration object.

    Returns:
        list: List of arguments to be used when executing Behave.
    """
    arguments = []
    output_folder = config.get_env('OUTPUT') if config else get_env('OUTPUT')
    if multiprocess:
        updated_features_path = features_path if not feature else feature
        updated_features_path = updated_features_path if not scenario_line else "{}:{}".format(updated_features_path, scenario_line)
        arguments.append(updated_features_path)
        arguments.append('--no-summary')
        worker_id = multiprocessing.current_process().name.split('-')[-1]

        # Note: --outfile removed since we get execution results directly from runner

        arguments.append('-D')
        arguments.append(f'worker_id={worker_id}')
    else:
        if type(features_path) is list:
            for feature_path in features_path:
                arguments.append(feature_path)
        else:
            arguments.append(features_path)
        if get_param('dry_run'):
            arguments.append('--no-summary')
            arguments.append('--dry-run')
        else:
            arguments.append('--summary')
        arguments.append('--junit-directory')
        arguments.append(output_folder)
        # Note: --outfile removed since we get execution results directly from runner
        arguments.append('-D')
        arguments.append(f'worker_id=0')
    arguments.append('--no-skipped')
    arguments.append('--no-junit')
    run_wip_tests = False
    env_tags = get_env('tags')
    if env_tags:
        tags = env_tags.split(';')
        for tag in tags:
            arguments.append('--tags')
            arguments.append(tag)
            if tag.upper() in ['WIP', '@WIP']:
                run_wip_tests = True
    if not run_wip_tests:
        arguments.append('--tags')
        arguments.append('~@WIP')
    arguments.append('--tags')
    arguments.append('~@MANUAL')

    # Handle output suppression based on execution mode
    if multiprocess:
        # In multiprocess: use null format to suppress verbose step output
        arguments.append('--format')
        arguments.append('null')
    else:
        # In single process: redirect output to behave.log to keep console clean
        output_folder = get_env('OUTPUT')
        if output_folder:
            behave_log_path = os.path.join(output_folder, 'behave', 'behave.log')
            # Ensure behave directory exists
            behave_dir = os.path.dirname(behave_log_path)
            if not os.path.exists(behave_dir):
                os.makedirs(behave_dir)
            arguments.append('--outfile')
            arguments.append(behave_log_path)

    args_sys = config.args if config else None
    set_args_captures(arguments, args_sys)
    if args_sys and args_sys.no_snippets:
        arguments.append('--no-snippets')
    for arg in BEHAVE_ARGS:
        value_arg = getattr(args_sys, arg) if hasattr(args_sys, arg) else False
        if arg == 'include':
            if multiprocess or not value_arg:
                continue
            else:
                features_path = os.path.abspath(os.environ['FEATURES_PATH'])
                value_arg = value_arg.replace(features_path, 'features').replace(
                    '\\', '\\\\'
                )
        if arg == 'define' and value_arg:
            for key_value in value_arg:
                arguments.append('--define')
                arguments.append(key_value)
        if value_arg and arg not in BEHAVEX_ARGS and arg != 'define':
            arguments.append('--{}'.format(arg.replace('_', '-')))
            if value_arg and not isinstance(value_arg, bool):
                arguments.append(value_arg)
                if arguments == 'logging_level':
                    set_env_variable(arg, value_arg)
                else:
                    os.environ[arg] = str(value_arg)
    return arguments


def _get_feature_json_skeleton(behave_element):
    """
    Get the JSON skeleton for the given feature or scenario.

    Args:
        behave_element (Feature or Scenario): Behave element.

    Returns:
        str: JSON skeleton of the feature.
    """
    if type(behave_element) is Feature:
        feature = behave_element
    elif type(behave_element) is Scenario:
        feature = copy.copy(behave_element.feature)
        feature.scenarios = [behave_element]
    else:
        raise Exception("No feature or scenario to process...")
    execution_info = generate_execution_info([feature])
    return json.dumps(execution_info[0]) if execution_info else {}


def _get_progress_bar_instance(parallel_scheme, total_elements):
    """
    Get the progress bar instance.

    Args:
        parallel_scheme (str): Parallel scheme (feature or scenario).
        total_elements (int): Total number of elements.

    Returns:
        ProgressBar: Progress bar instance.
    """
    try:
        config = conf_mgr.get_config()
        print_updates_in_new_lines = True if config['progress_bar']['print_updates_in_new_lines'] else False
        progress_bar_prefix = "Executed {}s".format(parallel_scheme)
        progress_bar_instance = ProgressBar(progress_bar_prefix,
                                            total_elements,
                                            print_updates_in_new_lines=print_updates_in_new_lines)
    except Exception as ex:
        print("There was an error creating the progress bar: {}".format(ex))
        return None
    return progress_bar_instance


def set_args_captures(args, args_sys):
    """
    Set the capture arguments for Behave.

    Args:
        args (list): List of arguments.
        args_sys (Namespace): Parsed command-line arguments.
    """
    for default_arg in ['capture', 'capture_stderr', 'logcapture']:
        if not getattr(args_sys, 'no_{}'.format(default_arg)):
            args.append('--no-{}'.format(default_arg.replace('_', '-')))



if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""
/*
* BehaveX - Agile test wrapper on top of Behave (BDD)
*/

JSON test execution report.
"""

# pylint: disable=W0212
# Future added in order to maintain compatibility
from __future__ import absolute_import

import json
import multiprocessing
import os
import random
import traceback
from tempfile import gettempdir

from behave.model import ScenarioOutline
from behave.step_registry import registry

from behavex.conf_mgr import get_env, get_param
from behavex.global_vars import global_vars
from behavex.outputs.report_utils import (get_environment_details,
                                          get_error_message, get_string_hash,
                                          match_for_execution, text)
from behavex.utils import (generate_hash, generate_uuid, get_scenario_tags,
                           retry_file_operation)


def add_step_info(step, parent_node):
    index = len(parent_node)
    parent_node.append(_step_to_dict(index, step))


def add_step_info_background(step, parent_node):
    step_info = {}
    for attrib in ('step_type', 'name', 'text', 'status'):
        step_info[attrib] = text(getattr(step, attrib))
    step_info['duration'] = step.duration or 0
    # Add start and stop times if they exist
    if hasattr(step, 'start'):
        step_info['start'] = getattr(step, 'start')
    if hasattr(step, 'stop'):
        step_info['stop'] = getattr(step, 'stop')
    if step.table:
        step_info['table'] = {}
        for heading in step.table.headings:
            step_info['table'][heading] = []
            for row in step.table:
                step_info['table'][heading].append(row[heading])
    step_info['index'] = len(parent_node)
    step_info['background'] = 'True'
    process_step_definition(step, step_info)
    parent_node.append(step_info)
    return step_info


def generate_execution_info(features):
    # Generate scenario list
    feature_list = []
    for feature in features:
        scenario_list = []
        id_feature = random.getrandbits(16)
        for feature_scenario in feature.scenarios:
            if isinstance(feature_scenario, ScenarioOutline):
                scenarios = feature_scenario.scenarios
            else:
                scenarios = [feature_scenario]
            scenario_list = _processing_scenarios(scenarios, scenario_list, id_feature)[1]

        if scenario_list:
            feature_info = {}
            for attrib in ('name', 'status', 'duration'):
                value = getattr(feature, attrib)
                value = value.name if attrib == 'status' else value
                # Override feature status to 'untested' during dry runs
                if attrib == 'status' and get_param('dry_run'):
                    value = 'untested'
                feature_info[attrib] = value
            feature_info['filename'] = text(feature.filename)
            feature_info['scenarios'] = scenario_list
            feature_info['background'] = _processing_background_feature(feature)
            feature_info['id'] = id_feature
            feature_list.append(feature_info)
    return feature_list


def generate_json_report(feature_list):
    output = {
        'environment': get_environment_details(),
        'features': feature_list,
        'steps_definition': global_vars.steps_definitions,
    }
    if multiprocessing.current_process().name == 'MainProcess':
        path_info = os.path.join(
            os.path.abspath(get_env('OUTPUT')),
            global_vars.report_filenames['report_json'],
        )
    else:
        path_info = os.path.join(
            gettempdir(),
            'result{}.tmp'.format(
                multiprocessing.current_process().name.split('-')[-1]
            ),
        )
    try:
        with open(path_info, 'w') as file_info:

            def write_json():
                file_info.write(json.dumps(output))

            retry_file_operation(path_info, execution=write_json)
        if multiprocessing.current_process().name != 'MainProcess':
            return output
    except IOError:
        msg = (
            'The file {0} is apparently being used. Please, close it and '
            'then try again'
        )
        raise Exception(msg.format(global_vars.report_filenames['report_json']))
    except Exception as exc_json:
        msg = (
            'An error occurred trying to generate_gallery file report.json '
            'error:{0} json:{1}'
        )
        raise Exception(msg.format(exc_json, output))


# noinspection PyBroadException
def _processing_background(scenario):
    scenario_background = {'duration': 0.0, 'steps': []}
    if scenario.background:
        steps = []
        for step in scenario._background_steps:
            step_back = add_step_info_background(step, steps)
            scenario_background['duration'] += step_back['duration']
        scenario_background['name'] = str(scenario.background.name)
        scenario_background['steps'] = steps
    return scenario_background


# noinspection PyBroadException
def _processing_background_feature(feature):
    feature_background = {}
    if feature.background:
        steps = []
        for step in feature.background.steps:
            add_step_info_background(step, steps)
        feature_background['name'] = str(feature.background.name)
        feature_background['steps'] = steps
    return feature_background


def _processing_scenarios(scenarios, scenario_list, id_feature):
    scenario_outline_index = 0
    overall_status = 'passed'
    is_dry_run = get_param('dry_run')
    for scenario in scenarios:
        # Remove BHX_MANUAL_DRY_RUN tag if it is a dry run
        scenario_tags = get_scenario_tags(scenario)
        if is_dry_run and 'BHX_MANUAL_DRY_RUN' in scenario_tags:
            scenario.tags.remove('BHX_MANUAL_DRY_RUN')
        # Set MANUAL to False in order filter regardless of it
        error_msg, error_lines, error_step, error_background = _get_error_scenario(
            scenario
        )
        # pylint: disable=W0123
        if match_for_execution(scenario_tags):
            # Scenario was selectable
            scenario_info = {}
            scenario_info['line'] = getattr(scenario, 'line')
            scenario_info['name'] = getattr(scenario, 'name')
            scenario_info['duration'] = getattr(scenario, 'duration')
            # Use the original status from behave
            original_status = getattr(scenario, 'status').name
            # In dry runs, treat failed and error scenarios as skipped since they weren't actually executed
            if is_dry_run and original_status in ['failed', 'error']:
                scenario_info['status'] = 'skipped'
            else:
                scenario_info['status'] = original_status
            # Add start and stop times if they exist
            if hasattr(scenario, 'start'):
                scenario_info['start'] = getattr(scenario, 'start')
            if hasattr(scenario, 'stop'):
                scenario_info['stop'] = getattr(scenario, 'stop')
            # Convert tags to list for JSON serialization (behave 1.3.0 returns set)
            tags = getattr(scenario, 'effective_tags')
            scenario_info['tags'] = list(tags) if isinstance(tags, set) else tags
            scenario_info['filename'] = text(scenario.filename)
            scenario_info['feature'] = scenario.feature.name
            scenario_info['id_feature'] = id_feature

            # Extract parameters for scenario outlines
            if hasattr(scenario, '_row') and scenario._row:
                scenario_info['parameters'] = dict(zip(scenario._row.headings, scenario._row.cells))

            steps = []
            for step in scenario.steps:
                add_step_info(step, steps)
            scenario_info['steps'] = steps
            scenario_info['outline_index'] = scenario_outline_index
            if scenario_info['status'] in ['failed', 'error']:
                # Prioritize error status over failed status for features
                if scenario_info['status'] == 'error':
                    overall_status = 'error'
                elif overall_status not in ['error']:  # Only set to failed if not already error
                    overall_status = 'failed'
            scenario_outline_index += 1
            scenario_info['background'] = _processing_background(scenario)
            scenario_info['error_msg'] = error_msg
            scenario_info['error_lines'] = error_lines
            scenario_info['error_step'] = error_step
            scenario_info['error_background'] = error_background
            scenario_info['id_hash'] = generate_uuid()
            scenario_info['identifier_hash'] = scenario.identifier_hash if hasattr(scenario, 'identifier_hash') else get_string_hash(f"{str(scenario.feature.filename)}-{str(scenario.line)}")
            scenario_info['process_id'] = str(scenario.process_id) if hasattr(scenario, 'process_id') else str(os.getpid())
            scenario_info['worker_id'] = str(scenario.worker_id) if hasattr(scenario, 'worker_id') else '0'
            if scenario.feature.name in global_vars.retried_scenarios:
                if (
                    scenario.name
                    in global_vars.retried_scenarios[scenario.feature.name]
                ):
                    scenario_info['retried'] = True

            scenario_list.append(scenario_info)
    return overall_status, scenario_list


def _get_error_scenario(scenario):
    error_lines = []
    error_msg = []
    failing_step = None
    error_background = False
    b_steps = scenario._background_steps if scenario._background_steps else []
    for index, step in enumerate(b_steps):
        if step.status == 'undefined' or (hasattr(step, 'exception') and step.exception):
            failing_step = _step_to_dict(index, step)
            failing_step['background'] = 'True'
            # failing_step.keys() is forced to be a list in order to maintain compatibility
            if 'error_msg' in list(failing_step.keys()):
                error_msg = failing_step['error_msg'].splitlines()
            if 'error_lines' in list(failing_step.keys()):
                error_lines = failing_step['error_lines']
            if step.status == 'undefined':
                error_msg = u'step undefined'
                error_lines = []
            error_background = True
            break
    for index, step in enumerate(scenario.steps):
        if hasattr(step, 'exception') and step.exception:
            failing_step = _step_to_dict(index, step)
            # failing_step.keys() is forced to be a list in order to maintain compatibility
            if 'error_msg' in list(failing_step.keys()):
                error_msg = failing_step['error_msg'].splitlines()
            if 'error_lines' in list(failing_step.keys()):
                error_lines = failing_step['error_lines']
            error_background = False
            break
    return error_msg, error_lines, failing_step, error_background


def _step_to_dict(index, step):
    step_info = {}
    for attrib in ('step_type', 'name', 'text', 'status'):
        step_info[attrib] = (
            text(getattr(step, attrib)) if text(getattr(step, attrib)) else ''
        )
    step_info['duration'] = 0.0
    if hasattr(step, 'duration'):
        step_info['duration'] = step.duration or 0.0
    else:
        step_info['duration'] = 0.0
    if hasattr(step, 'line'):
        step_info['line'] = getattr(step, 'line')
    # Add start and stop times if they exist
    if hasattr(step, 'start'):
        step_info['start'] = getattr(step, 'start')
    if hasattr(step, 'stop'):
        step_info['stop'] = getattr(step, 'stop')
    if hasattr(step, 'exception') and step.exception:
        # step.exception is forced to be a str type variable
        step_info['error_msg'] = get_error_message(str(step.exception))
        # Handle traceback formatting safely
        try:
            # Check if exception has a traceback attribute
            if hasattr(step.exception, '__traceback__'):
                step_info['error_lines'] = traceback.format_exception(
                    type(step.exception), step.exception, step.exception.__traceback__)
            else:
                # Fallback for older Python versions or exceptions without traceback
                step_info['error_lines'] = traceback.format_exc().splitlines()
        except Exception:
            # Provide a safe fallback if traceback extraction fails
            step_info['error_lines'] = get_error_message(str(step.exception))
    if step.table:
        step_info['table'] = {}
        for heading in step.table.headings:
            step_info['table'][heading] = []
            for row in step.table:
                step_info['table'][heading].append(row[heading])
    process_step_definition(step, step_info)
    step_info['index'] = index
    return step_info


def process_step_definition(step, step_info):
    definition = registry.find_step_definition(step)
    if definition:
        hash_step = generate_hash(definition.pattern)
        if hash_step not in global_vars.steps_definitions:
            global_vars.steps_definitions[hash_step] = definition.pattern
        step_info['hash'] = hash_step
    else:
        step_info['hash'] = 0



import os
import shutil
import subprocess
import tempfile

from behave import given, then, when
from bs4 import BeautifulSoup

from behavex.runner import run


@given('I have created feature files with same tags but different names')
def step_create_feature_files(context):
    # Create temporary directory for test features
    context.temp_dir = tempfile.mkdtemp()
    context.output_dir = tempfile.mkdtemp()

    # Create feature files with same tag but different names
    feature_1_content = """@test_A
Feature: Test A part 1

Background:
  Given some common setup for Test A part 1

@scenario11
Scenario: Scenario 11
  Given I am on the first part page
  When I click on the first button
  Then I should see the first result

@scenario12
Scenario: Scenario 12
  Given I am on the first part page
  When I click on the second button
  Then I should see the second result
"""

    feature_2_content = """@test_A
Feature: Test A part 2

@scenario21
Scenario: Scenario 21
  Given I am on the second part page
  When I click on the first button
  Then I should see the third result

@scenario22
Scenario: Scenario 22
  Given I am on the second part page
  When I click on the second button
  Then I should see the fourth result
"""

    feature_3_content = """@test_A
Feature: Test A part 3

@scenario31
Scenario: Scenario 31
  Given I am on the third part page
  When I click on the first button
  Then I should see the fifth result

@scenario32
Scenario: Scenario 32
  Given I am on the third part page
  When I click on the second button
  Then I should see the sixth result
"""

    # Write feature files
    with open(os.path.join(context.temp_dir, 'test_a_part_1.feature'), 'w') as f:
        f.write(feature_1_content)

    with open(os.path.join(context.temp_dir, 'test_a_part_2.feature'), 'w') as f:
        f.write(feature_2_content)

    with open(os.path.join(context.temp_dir, 'test_a_part_3.feature'), 'w') as f:
        f.write(feature_3_content)

    # Create steps directory and step definitions
    steps_dir = os.path.join(context.temp_dir, 'steps')
    os.makedirs(steps_dir)

    steps_content = """from behave import given, when, then

@given('some common setup for Test A part 1')
def step_impl(context):
    pass

@given('I am on the first part page')
def step_impl(context):
    pass

@given('I am on the second part page')
def step_impl(context):
    pass

@given('I am on the third part page')
def step_impl(context):
    pass

@when('I click on the first button')
def step_impl(context):
    pass

@when('I click on the second button')
def step_impl(context):
    pass

@then('I should see the first result')
def step_impl(context):
    pass

@then('I should see the second result')
def step_impl(context):
    pass

@then('I should see the third result')
def step_impl(context):
    pass

@then('I should see the fourth result')
def step_impl(context):
    pass

@then('I should see the fifth result')
def step_impl(context):
    pass

@then('I should see the sixth result')
def step_impl(context):
    pass
"""

    with open(os.path.join(steps_dir, '__init__.py'), 'w') as f:
        f.write('# Step definitions for feature nesting test')

    with open(os.path.join(steps_dir, 'nesting_test_steps.py'), 'w') as f:
        f.write(steps_content)


@when('I run BehaveX on these feature files with parallel scenario execution')
def step_run_parallel_scenario(context):
    context.execution_mode = 'parallel_scenario'
    _run_behavex(context, ['--parallel-processes=2', '--parallel-scheme=scenario'])


@when('I run BehaveX on these feature files with parallel feature execution')
def step_run_parallel_feature(context):
    context.execution_mode = 'parallel_feature'
    _run_behavex(context, ['--parallel-processes=2', '--parallel-scheme=feature'])


@when('I run BehaveX on these feature files with single process execution')
def step_run_single_process(context):
    context.execution_mode = 'single_process'
    _run_behavex(context, ['--parallel-processes=1'])


def _run_behavex(context, extra_args):
    # Run BehaveX with the specified arguments
    cmd = ['python', '-m', 'behavex', context.temp_dir, '-o', context.output_dir] + extra_args
    result = subprocess.run(cmd, capture_output=True, text=True, cwd=context.temp_dir)
    context.behavex_result = result
    context.html_report_path = os.path.join(context.output_dir, 'report.html')


@then('I should see features displayed flat in the HTML report')
def step_check_flat_display(context):
    # Parse the HTML report and check structure
    assert os.path.exists(context.html_report_path), f"HTML report not found at {context.html_report_path}"

    with open(context.html_report_path, 'r') as f:
        html_content = f.read()

    soup = BeautifulSoup(html_content, 'html.parser')

    # Find all feature rows (these should be at the same level)
    feature_rows = soup.find_all('tr', {'class': lambda x: x and 'data_feature' in x})

    # Should have exactly 3 feature rows
    assert len(feature_rows) == 3, f"Expected 3 feature rows, found {len(feature_rows)}"

    # Extract feature names
    feature_names = []
    for row in feature_rows:
        feature_name_span = row.find('span', {'data-feature': True})
        if feature_name_span:
            feature_name = feature_name_span.find('b').text.strip()
            feature_names.append(feature_name)

    context.feature_names = feature_names

    # Should contain all expected feature names
    expected_names = ['Test A part 1', 'Test A part 2', 'Test A part 3']
    for name in expected_names:
        assert name in feature_names, f"Feature '{name}' not found in feature list: {feature_names}"


@then('I should not see features nested within other features')
def step_check_no_nesting(context):
    # Check that no feature rows are nested inside other feature rows
    with open(context.html_report_path, 'r') as f:
        html_content = f.read()

    soup = BeautifulSoup(html_content, 'html.parser')

    # Find all feature containers
    feature_containers = soup.find_all('tr', {'data-feature-id': True})

    # Check that none of these containers contain other feature rows
    for container in feature_containers:
        nested_features = container.find_all('tr', {'class': lambda x: x and 'data_feature' in x})
        # A container might contain the feature row itself, but not additional feature rows
        assert len(nested_features) <= 1, f"Found nested features within a feature container: {[f.text for f in nested_features]}"


@then('each feature should have its own row in the summary table')
def step_check_summary_table(context):
    # Verify that each feature has its own row in the main summary table
    with open(context.html_report_path, 'r') as f:
        html_content = f.read()

    soup = BeautifulSoup(html_content, 'html.parser')

    # Find the summary table
    summary_table = soup.find('table', {'id': 'summary'})
    assert summary_table is not None, "Summary table not found"

    # Count feature rows in summary table
    feature_rows = summary_table.find_all('tr', {'class': lambda x: x and 'data_feature' in x})
    assert len(feature_rows) == 3, f"Expected 3 feature rows in summary table, found {len(feature_rows)}"

    # Clean up temp directories
    if hasattr(context, 'temp_dir'):
        shutil.rmtree(context.temp_dir, ignore_errors=True)
    if hasattr(context, 'output_dir'):
        shutil.rmtree(context.output_dir, ignore_errors=True)




def before_all(context):
    # Configure behavex-images to always attach images to reports
    try:
        from behavex_images import image_attachments
        from behavex_images.image_attachments import AttachmentsCondition
        image_attachments.set_attachments_condition(context, AttachmentsCondition.ALWAYS)
    except (ImportError, ModuleNotFoundError):
        pass  # behavex-images not available


def after_scenario(context, scenario):
    if hasattr(context, 'new_scenario_name'):
        scenario.name = context.new_scenario_name
    if hasattr(context, 'new_feature_name'):
        scenario.feature.name = context.new_feature_name


def after_feature(context, feature):
    """Ensure after_feature hook is called by behave (required for behave 1.2.7+ compatibility)"""
    # Call BehavEx after_feature directly since hook interception doesn't receive the feature object
    try:
        from behavex import environment as behavex_env
        behavex_env.after_feature(context, feature)
    except Exception:
        pass  # Ignore errors to avoid breaking test execution



"""
/*
* BehaveX - Agile test wrapper on top of Behave (BDD)
*/

Formatter manager module to handle custom formatters.
"""

import importlib
import os
import sys
from typing import (Any, Dict, Optional,  # pyright: ignore[reportDeprecated]
                    Type)

from behavex.conf_mgr import get_env, get_param

# Formatter Constants
DEFAULT_FORMATTER_DIR = 'report-artifacts'


class FormatterManager:
    """Manager class for handling custom formatters in BehaveX."""

    @staticmethod
    def load_formatter(formatter_spec: str) -> Optional[Type[Any]]: #pyright: ignore[reportDeprecated]
        """
        Load a formatter class from a module path specification.

        Args:
            formatter_spec: String in format "module_path:FormatterClass"

        Returns:
            The formatter class if found, None otherwise
        """
        try:
            if not formatter_spec or ':' not in formatter_spec:
                print("Invalid formatter specification. Format should be 'module_path:FormatterClass'")
                return None

            # Clean up the formatter spec by removing any extra whitespace
            formatter_spec = formatter_spec.strip()
            module_path, class_name = [part.strip() for part in formatter_spec.split(':')]

            # Try to import the module
            try:
                module = importlib.import_module(module_path)
            except ImportError as e:
                # If import fails, try to add the current directory to sys.path
                current_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
                if current_dir not in sys.path:
                    sys.path.append(current_dir)
                try:
                    module = importlib.import_module(module_path)
                except ImportError:
                    print(f"Could not import module '{module_path}'. Error: {str(e)}")
                    return None

            # Get the formatter class
            if not hasattr(module, class_name):
                print(f"Module '{module_path}' does not contain class '{class_name}'")
                return None

            formatter_class = getattr(module, class_name)
            return formatter_class

        except Exception as e:
            print(f"Error loading formatter {formatter_spec}: {str(e)}")
            return None

    @staticmethod
    def get_formatter_output_dir(formatter_spec: str) -> Optional[str]: #pyright: ignore[reportDeprecated]
        """
        Get the preferred output directory from a formatter class.

        Args:
            formatter_spec: String in format "module_path:FormatterClass"

        Returns:
            The formatter's preferred output directory, or None if not found
        """
        try:
            formatter_class = FormatterManager.load_formatter(formatter_spec)
            if formatter_class and hasattr(formatter_class, 'DEFAULT_OUTPUT_DIR'):
                return formatter_class.DEFAULT_OUTPUT_DIR
        except Exception as e:
            print(f"Warning: Could not load formatter {formatter_spec} to get output directory: {str(e)}")

        return None

    @staticmethod
    def format_results(json_output: Dict[str, Any]) -> None: #pyright: ignore[reportDeprecated]
        """
        Format test results using the specified custom formatter.

        Args:
            json_output: Dictionary containing the test results
        """
        formatter_spec = get_param('formatter')
        if not formatter_spec:
            return

        formatter_class = FormatterManager.load_formatter(formatter_spec)
        if not formatter_class:
            return

        # The formatter will handle its own output directory setup
        # No need to create directories here since formatters manage their own paths

        try:
            formatter = formatter_class()
            # Check if the formatter has the expected method
            if not hasattr(formatter, 'launch_json_formatter'):
                print(f"Formatter {formatter_class.__name__} does not implement required method 'launch_json_formatter'")
                return

            formatter.launch_json_formatter(json_output)
        except Exception as e:
            print(f"Error running formatter {formatter_spec}: {str(e)}")
            import traceback
            traceback.print_exc()



import logging
import os

from behave import given, then, when

# Get paths
root_project_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..', '..', '..'))
test_images_path = os.path.join(root_project_path, 'tests', 'test_images')

# Import behavex-images functionality - fail if required
def get_image_attachments_module():
    """Get the image_attachments module, raising an error if not available."""
    try:
        from behavex_images import image_attachments
        return image_attachments
    except (ImportError, ModuleNotFoundError) as e:
        raise AssertionError(f"behavex-images is required for image attachment tests but not available: {e}")

# Check if behavex-images is available for optional usage
try:
    from behavex_images import image_attachments
    behavex_images_available = True
except (ImportError, ModuleNotFoundError) as e:
    behavex_images_available = False
    logging.warning(f"behavex-images not available, image attachment tests will be skipped: {e}")


@given('a failing condition')
def given_failing_condition(context):
    context.condition = 'fail'
    logging.info('a failing condition')

@given('a passing condition')
def given_passing_condition(context):
    context.condition = 'pass'
    logging.info('a passing condition')

@given('image attachments are set to ONLY_ON_FAILURE condition')
def given_image_attachments_only_on_failure(context):
    """Set image attachments to ONLY_ON_FAILURE condition"""
    try:
        from behavex_images import image_attachments
        from behavex_images.image_attachments import AttachmentsCondition
        image_attachments.set_attachments_condition(context, AttachmentsCondition.ONLY_ON_FAILURE)
        logging.info('Image attachments set to ONLY_ON_FAILURE condition')
    except (ImportError, ModuleNotFoundError):
        logging.warning('behavex-images not available, skipping attachment condition setup')

@then('a failing condition is performed')
def then_failing_condition_performed(context):
    """Perform a failing assertion to create failed status"""
    logging.info('Performing failing condition - this will cause failed status')
    assert False, "This step is designed to fail"



@given('a passing condition that records execution order "{order_tag}"')
def given_passing_condition_with_order(context, order_tag):
    context.condition = 'pass'
    # Record execution order for testing purposes
    if not hasattr(context, 'execution_order'):
        context.execution_order = []
    context.execution_order.append(order_tag)
    logging.info(f'a passing condition that records execution order "{order_tag}"')

@given('a condition to skip the scenario')
def given_skip_condition(context):
    context.condition = 'skip'
    logging.info('a condition to skip the scenario')

@given('a condition to exit the scenario')
def given_exit_condition(context):
    context.condition = 'exit'
    logging.info('a condition to exit the scenario')

@given('a condition to leave the scenario untested')
def given_untested_condition(context):
    context.condition = 'untested'
    logging.info('a condition to leave the scenario untested')

@then('I perform the condition')
def then_perform_condition(context):
    if context.condition == 'fail':
        # This step will cause the test to fail
        assert False, "This step is designed to fail"
    elif context.condition == 'pass':
        # This step will pass
        assert True
    elif context.condition == 'skip':
        # This step will be skipped
        context.scenario.skip("This scenario is skipped")
    elif context.condition == 'exit':
        # This step will be skipped
        exit(1)
    elif context.condition == 'untested':
        # This step will be skipped
        pass

@given('I rename the {feature_or_scenario} from context to have the suffix "{suffix}"')
def given_rename_feature_or_scenario(context, feature_or_scenario, suffix):
    if feature_or_scenario == 'feature':
        context.new_feature_name = context.feature.name + suffix
        logging.info('I rename the feature from \n"{}" \nto \n"{}"'.format(context.feature.name, context.new_feature_name))
    elif feature_or_scenario == 'scenario':
        context.new_scenario_name = context.scenario.name + suffix
        logging.info('I rename the scenario from \n"{}" \nto \n"{}"'.format(context.scenario.name, context.new_scenario_name))
    else:
        raise ValueError('Invalid element, it should be "feature" or "scenario"')


@given('I take a screenshot using test image {image_number}')
def step_take_screenshot_using_test_image(context, image_number):
    """Attach a test image file using behavex-images"""
    image_path = os.path.join(test_images_path, f'test_image_{image_number}.png')
    if not os.path.exists(image_path):
        raise FileNotFoundError(f"Test image not found: {image_path}")

    # Check if image attachments are required for this test
    if hasattr(context, 'require_image_attachments') and context.require_image_attachments:
        image_attachments = get_image_attachments_module()
        image_attachments.attach_image_file(context, image_path)
        logging.info(f"Attached test image: {image_path}")
    else:
        # Optional usage - skip if not available
        if not behavex_images_available:
            logging.warning("behavex-images not available, skipping image attachment")
            return
        # Import here to avoid issues if behavex-images is not available
        from behavex_images import image_attachments
        image_attachments.attach_image_file(context, image_path)
        logging.info(f"Attached test image: {image_path}")


@when('I validate image attachments')
def step_validate_image_attachments(context):
    """Simple validation step for image attachments"""
    logging.info("Validating image attachments")


@then('I should see attached images in the output')
def step_verify_attached_images_in_output(context):
    """Verify that images were attached"""
    logging.info("Images were attached to the test output")


# Simple step definitions for test subject feature files
@given('a simple test condition')
def step_simple_test_condition(context):
    """Simple test condition for test subject scenarios"""
    context.test_condition = True
    logging.info("Simple test condition established")


@when('I perform a simple action')
def step_perform_simple_action(context):
    """Simple action for test subject scenarios"""
    context.action_performed = True
    logging.info("Simple action performed")


@then('I should see a simple result')
def step_simple_result(context):
    """Simple result verification for test subject scenarios"""
    assert context.test_condition, "Test condition was not established"
    assert context.action_performed, "Action was not performed"
    logging.info("Simple result verified")


# Step definitions for allure_text_tests.feature
@given('a test condition with multiline text')
def step_test_condition_with_multiline_text(context):
    """Test condition with multiline text"""
    context.multiline_text_condition = True
    logging.info("Test condition with multiline text established")


@when('I process the following text:')
def step_process_multiline_text(context):
    """Process multiline text"""
    context.processed_text = context.text
    context.text_processing_action = True
    logging.info("Multiline text processed")


@then('I should see the text processed correctly')
def step_see_text_processed_correctly(context):
    """Verify text processed correctly"""
    assert context.multiline_text_condition, "Multiline text condition was not established"
    assert context.text_processing_action, "Text processing action was not executed"
    assert context.processed_text, "Text was not processed"
    logging.info("Text processing verified")


# Step definitions for allure_table_tests.feature
@given('a test condition with table data')
def step_test_condition_with_table_data(context):
    """Test condition with table data"""
    context.table_data_condition = True
    logging.info("Test condition with table data established")


@when('I process the following table:')
def step_process_table_data(context):
    """Process table data"""
    context.processed_table = context.table
    context.table_processing_action = True
    logging.info("Table data processed")


@then('I should see the table processed correctly')
def step_see_table_processed_correctly(context):
    """Verify table processed correctly"""
    assert context.table_data_condition, "Table data condition was not established"
    assert context.table_processing_action, "Table processing action was not executed"
    assert context.processed_table, "Table was not processed"
    logging.info("Table processing verified")


# Step definitions for allure_tagged_tests.feature
@given('a test condition with allure tags')
def step_test_condition_with_allure_tags(context):
    """Test condition with allure tags"""
    context.allure_tags_condition = True
    logging.info("Test condition with allure tags established")


@given('another test condition with different tags')
def step_another_test_condition_with_different_tags(context):
    """Another test condition with different tags"""
    context.different_tags_condition = True
    logging.info("Another test condition with different tags established")


@given('a critical test condition')
def step_critical_test_condition(context):
    """Critical test condition"""
    context.critical_condition = True
    logging.info("Critical test condition established")


@when('I execute an action with epic and story tags')
def step_execute_action_with_epic_story_tags(context):
    """Execute action with epic and story tags"""
    context.epic_story_action = True
    logging.info("Action with epic and story tags executed")


@when('I execute another action')
def step_execute_another_action(context):
    """Execute another action"""
    context.another_action = True
    logging.info("Another action executed")


@when('I execute a critical action')
def step_execute_critical_action(context):
    """Execute a critical action"""
    context.critical_action = True
    logging.info("Critical action executed")


@then('I should see the result processed correctly')
def step_see_result_processed_correctly(context):
    """Verify result processed correctly"""
    assert context.allure_tags_condition, "Allure tags condition was not established"
    assert context.epic_story_action, "Epic story action was not executed"
    logging.info("Result processed correctly")


@then('I should see different tag processing')
def step_see_different_tag_processing(context):
    """Verify different tag processing"""
    assert context.different_tags_condition, "Different tags condition was not established"
    assert context.another_action, "Another action was not executed"
    logging.info("Different tag processing verified")


@then('I should see critical processing')
def step_see_critical_processing(context):
    """Verify critical processing"""
    assert context.critical_condition, "Critical condition was not established"
    assert context.critical_action, "Critical action was not executed"
    logging.info("Critical processing verified")



# -*- coding: utf-8 -*-
import os
import time


class GlobalVars:
    def __init__(self):
        # Ensure BEHAVEX_PATH is set if not already
        if 'BEHAVEX_PATH' not in os.environ:
            os.environ['BEHAVEX_PATH'] = os.path.dirname(os.path.realpath(__file__))
        self._execution_path = os.environ.get('BEHAVEX_PATH')
        self._report_filenames = {
            'report_json': 'report.json',
            'report_overall': 'overall_status.json',
            'report_failures': 'failing_scenarios.txt',
        }
        self._behave_tags_file = os.path.join('behave', 'behave.tags')
        self._jinja_templates_path = os.path.join(
            self._execution_path, 'outputs', 'jinja'
        )
        self._jinja_templates = {
            'main': 'main.jinja2',
            'steps': 'steps.jinja2',
            'xml': 'xml.jinja2',
            'xml_json': 'xml_json.jinja2',
            'manifest': 'manifest.jinja2',
        }
        self._retried_scenarios = {}
        self._steps_definitions = {}
        self._rerun_failures = False
        self._progress_bar_instance = None
        self._execution_start_time = time.time()
        self._execution_end_time = None


    @property
    def execution_path(self):
        return self._execution_path

    @property
    def report_filenames(self):
        return self._report_filenames

    @property
    def behave_tags_file(self):
        return self._behave_tags_file

    @property
    def jinja_templates_path(self):
        return self._jinja_templates_path

    @property
    def jinja_templates(self):
        return self._jinja_templates

    @property
    def retried_scenarios(self):
        return self._retried_scenarios

    @retried_scenarios.setter
    def retried_scenarios(self, feature_name):
        self._retried_scenarios[feature_name] = []

    @property
    def steps_definitions(self):
        return self._steps_definitions

    @property
    def rerun_failures(self):
        return self._rerun_failures

    @rerun_failures.setter
    def rerun_failures(self, rerun_failures):
        self._rerun_failures = rerun_failures

    @property
    def progress_bar_instance(self):
        return self._progress_bar_instance

    @progress_bar_instance.setter
    def progress_bar_instance(self, progress_bar_instance):
        self._progress_bar_instance = progress_bar_instance

    @property
    def execution_start_time(self):
        return self._execution_start_time

    @execution_start_time.setter
    def execution_start_time(self, execution_start_time):
        self._execution_start_time = execution_start_time

    @property
    def execution_elapsed_time(self):
        return time.time() - self._execution_start_time

    @property
    def execution_end_time(self):
        if not self._execution_end_time:
            self._execution_end_time = time.time()
        return self._execution_end_time



global_vars = GlobalVars()
